from shapely.geometry import Point
import numpy as np
from itertools import chain
from pyvista import PolyData


def nodes_conn(obj):

    """
    Define the topology of a network using networkx.degree. With this method also censored fractures are defined and
    node origin are calculated. By node origin we define which entities are related to the given node e.g.:
        + I node with node origin [x] -> related to a fracture of set x
        + Y node with node origin [x, y] -> related to a fracture of set x abutting on a fracture of set y
        + Y node with node origin [x, x] -> related to the intersection of fractures of the same set x
        + U node with node origin [x, b] -> related to a fracture of set x intersecting the boundary (b)
        + Y node with node origin [x, y, z] -> triple intersection, makes no sense -> problem in the geometry
        + X node with node origin [x, y] -> related to the intersection of fractures of set x and y
        + X node with node origin [w, x, y, z] -> quadruple intersection, makes no sense -> problem in the geometry

    :param obj: Fractures or FractureNetwork object
    :return: list of shapely geometry points, a list of corresponding node classes and a list of node origin
    """

    fractures_vtk = obj.fractures.vtk_object
    boundary_vtk = obj.boundaries.vtk_object
    entity_df_obj = obj.fracture_network_to_components_df()
    point_dict = dict()
    origin_dict = dict()
    ids = []

    # to get all the nodes we append the start and end of a line (I nodes) and then the entire point_id list
    for i, cell in enumerate(fractures_vtk.cell):
        cell_ids = cell.point_ids
        ids.append([cell_ids[0]])
        ids.append([cell_ids[-1]])
        ids.append(cell_ids)

    # we then create a flattened list of the indexes
    ids = list(chain.from_iterable(ids))

    # and get the unique values that repeat more than once.
    # i.e. I nodes and all the nodes that repeat because of intersection with the fractures
    u, c = np.unique(ids, return_counts=True)

    ids = u[c > 1]

    # To define boundary intersection we search for overlapping points (with a decimal precision of e-5).
    # This method has a bit of a problem because vtk and shapely do not have the same precision and thus when comparing
    # points after the e-5 values are different
    fracture_points = fractures_vtk.points
    boundary_points = boundary_vtk.points

    boundary_common = np.isin(np.round(fracture_points, 5), np.round(boundary_points, 5)).all(axis=1)
    boundary_index = np.where(boundary_common == 1)[0]

    # After we explode the fractures_vtk to get all the segments of each line
    fractures_segm = fractures_vtk.extract_all_edges(use_all_points=True)

    # We then loop for each node id that we found at the start
    for i in ids:
        # this indicates the degree of the node (i.e. I, Y or X)
        neigh = fracture_points[fractures_segm.point_neighbors(i)]

        # these will be the cells id that contain the given node idx
        cells = fractures_vtk.point_cell_ids(i)

        # To get the node origin we find the unique values of the fracture set at the given node idx and sort
        # using argsort to get the order that we want (i.e. increasing frequency)
        sets = fractures_segm['f_set'][fractures_segm.point_cell_ids(i)]
        u_sets, f_sets = np.unique(sets, return_counts=True)
        sorting_index = np.argsort(f_sets)

        print(u_sets[sorting_index])

        if len(cells) >= 3:
            print(f'\n\nInvalid point for lines: {fractures_vtk["og_line_id"][cells]} '
                  f'\n\nsets: {fractures_vtk["f_set"][cells]}, \n\nThe node will be classified accordingly'
                  f' to the number of intersection however, the intersection must be checked!')

        if len(neigh) != 2:
            point = Point(fracture_points[i])
            point_dict[point] = [len(neigh), i]
            origin_dict[point] = f'{u_sets[sorting_index]}'

    for i in boundary_index:
        point = Point(fracture_points[i])
        f_cells = fractures_vtk.point_cell_ids(i)
        origin_set = f'{fractures_vtk["f_set"][f_cells][0]}'

        point_dict[point] = [5, i]
        origin_dict[point] = f'{[origin_set, "b"]}'

    censored_lines = [fractures_vtk.point_cell_ids(idx)[0] for idx in boundary_index]
    entity_df_obj.loc[censored_lines, 'censored'] = 1
    obj.entity_df = entity_df_obj
    #fracture_nodes = PolyData(fractures_vtk.points)
    #fracture_nodes['topology'] = fractures_vtk['topology']

    # network_obj = obj.network_object()
    #
    # vtk_obj = obj.vtk_object(include_nodes=False)
    #
    # entity_df_obj = obj.fracture_network_to_components_df()
    #
    # fractures_vtk_obj = obj.fractures.vtk_object
    #
    # frac_idx = np.where(vtk_obj.point_data['type'] == 'fracture')[0]
    #
    # node_dict = {}
    # tot_idx = len(frac_idx)
    # print('\n\n')
    # for node in frac_idx:  # For each node of the fractures:
    #     print(f'Analyzing nodes:{node+1}/{tot_idx}', end='\r')
    #
    #     n_edges = network_obj.degree[node]  # Calculate number of connected nodes
    #
    #     point = Point(vtk_obj.points[node])
    #
    #     cells = fractures_vtk_obj.extract_points(node)
    #     if cells.number_of_cells == 0:
    #         node_origin = -9999
    #     else:
    #         node_origin = ''.join([str(int(c)) for c in set(cells['f_set'])])
    #
    #     if n_edges == 2:  # Exclude internal and V nodes
    #         n_edges = -9999
    #
    #     elif n_edges == 3:  # Discriminate Y (3 or 6) and U (5) nodes
    #
    #         cells = vtk_obj.extract_points(node)
    #
    #         if 'boundary' in cells['type']:
    #
    #             n_edges = 5
    #             index = vtk_obj.point_data['RegionId'][node]
    #             entity_df_obj.loc[index, 'censored'] = 1
    #
    #         else:
    #             if len(node_origin) > 1:  # Discriminate between Y nodes between different sets (3) or same set (6)
    #                 n_edges = 6
    #             else:
    #                 n_edges = 3
    #
    #     elif n_edges == 4:
    #         cells = vtk_obj.extract_points(node)
    #         if 'boundary' in cells['type']:
    #             n_edges = 5
    #             index = vtk_obj.point_data['RegionId'][node]
    #             entity_df_obj.loc[index, 'censored'] = 1
    #         else:
    #             n_edges = 4
    #
    #     elif n_edges > 4:
    #
    #         n_edges = -9999
    #
    #     node_dict[point] = (n_edges, node_origin)
    #
    # obj.entity_df = entity_df_obj
    # print('\n\nDone!')
    return point_dict, origin_dict


# def find_backbone(obj: FractureNetwork) -> PolyData:
#
#     fractures = obj.fractures_components.vtk_object
#
#     connectivity = vtkConnectivityFilter()
#
#     connectivity.AddInputData(fractures)
#     connectivity.SetExtractionModeToLargestRegion()
#     connectivity.Update()
#
#     backbone = PolyData(connectivity.GetOutput())
#
#     return backbone



import scooby
import pyperclip
import numpy as np
import pyvista as pv
from geopandas import GeoDataFrame
from vtkmodules.vtkFiltersCore import vtkCleanPolyData


def report():
    """ Method used to create a report using scooby and copy it in the clipboard"""
    core = ['fracability', 'pyvista', 'vtk', 'numpy', 'geopandas', 'shapely']
    text = scooby.Report(core=core, ncol=3, text_width=80, sort=True, additional=None, optional=None)
    print(text)
    pyperclip.copy(str(text))
    print('Report copied to clipboard')


def centers_to_lines(center_coords, lengths, frac_dir, assign_id=True) -> pv.PolyData:
    """ Function used to create fractures of given lengths from center coordinates.

    :param center_coords: xyz coordinate centers array of the fractures
    :param lengths: length array for each fracture.
    :param frac_dir: direction array of each fracture.
    :param assign_id: Flag used to assign the id of each fracture as a field (RegionId). Default True

    :return: Pyvista polydata with the same number of fracture as centers.
    """

    half_lengths = lengths/2
    xyz1 = center_coords.copy()
    xyz2 = center_coords.copy()

    xyz1[:, 0] = center_coords[:, 0] + half_lengths * np.sin(np.deg2rad(frac_dir))
    xyz2[:, 0] = center_coords[:, 0] + half_lengths * np.sin(np.deg2rad((frac_dir + 180) % 360))

    xyz1[:, 1] = center_coords[:, 1] + half_lengths * np.cos(np.deg2rad(frac_dir))
    xyz2[:, 1] = center_coords[:, 1] + half_lengths * np.cos(np.deg2rad((frac_dir + 180) % 360))

    xyz_complete = np.array([[i, j] for i, j in zip(xyz1, xyz2)]).reshape(-1, 3)
    conn = np.insert(np.arange(0, len(xyz_complete)), np.arange(0, len(xyz_complete), 2), 2)
    lines = pv.PolyData(xyz_complete, lines=conn)

    if assign_id:
        regions_ids = np.arange(0, len(center_coords))
        lines['RegionId'] = regions_ids

    return lines


def KM(z_values, Z, delta_list):

    """
    Calculate the Kaplan-Meier curve given an input z, data Z and list of deltas.
    :param z_values: Input
    :param Z: Data (sorted)
    :param delta_list: list of deltas (sorted as Z)
    :return:
    """

    def p_cap(n, j_index, d_list):
        """
        Calculate the ^p estimator (formula 2.6)
        :param n: total number
        :param j_index: index list
        :param d_list: delta list
        :return:
        """
        product = 1  # initiate product as 1 (so that the first step p1 will be product=p1

        for j in j_index:  # for each index in the index list
            d = d_list[j]
            real_j = j + 1
            p = ((n - real_j) / (n - real_j + 1)) ** d
            product *= p

        return 1 - product

    # Sort Z in case it is not sorted at input (also delta_list needs to be sorted in the same order of Z)
    sorted_args = np.argsort(Z)
    Z_sort = Z[sorted_args]
    delta_list_sort = delta_list[sorted_args]

    G = np.ones_like(z_values)
    n = len(Z)
    for i, z in enumerate(z_values):
        if z < Z_sort[0]:
            G[i] = 0
        elif z <= Z_sort[-1]:
            j_index = np.where(Z_sort <= z)[0]  # Get the index in which the data Z is lower than z
            G[i] = p_cap(n, j_index, delta_list_sort)  # Calculate the p_cap
        elif z > Z_sort[-1]:
            G[i] = 1

    return G


def ecdf_find_x(samples: np.ndarray, ecdf_prob: np.ndarray, y_values: np.ndarray) -> list:
    """
    Find the corresponding sample value of the ecdf given an array of y values
    :param samples: Array of samples
    :param ecdf_prob: Array of ecdf values
    :param y_values: Array of y values to find the corresponding x
    :return: list of x values
    """

    x_values = []
    for y in y_values:
        mask = ecdf_prob < y
        last_true = np.count_nonzero(mask) - 1
        first_false = last_true + 1
        if last_true < 0:
            int_x = samples[0]
        elif first_false == len(ecdf_prob):
            int_x = samples[-1]
        else:
            int_x = np.mean([samples[last_true], samples[first_false]])

        x_values.append(np.round(int_x, 2))
    return x_values


def setAxLinesBW(ax):
    """
    Take each Line2D in the axes, ax, and convert the line style to be
    suitable for black and white viewing.
    Code taken from https://stackoverflow.com/questions/7358118/black-white-colormap-with-dashes-dots-etc
    """
    MARKERSIZE = 1

    COLORMAP = [{'marker': '', 'dash': (None, None)},
                {'marker': '', 'dash': [5, 5]},
                {'marker': '', 'dash': [5, 3, 1, 3]},
                {'marker': '', 'dash': [1, 3]},
                {'marker': '', 'dash': [5, 2, 5, 2, 5, 10]},
                {'marker': '', 'dash': [5, 3, 1, 2, 1, 10]},
                {'marker': '', 'dash': [1, 2, 1, 10]}
                ]

    lines_to_adjust = ax.get_lines()
    try:
        lines_to_adjust += ax.get_legend().get_lines()
    except AttributeError:
        pass

    for i, line in enumerate(lines_to_adjust):
        line.set_color('black')
        line.set_dashes(COLORMAP[i]['dash'])
        line.set_marker(COLORMAP[i]['marker'])
        line.set_markersize(MARKERSIZE)


def setFigLinesBW(fig):

    """
    Take each axes in the figure, and for each line in the axes, make the
    line viewable in black and white.
    Code taken from https://stackoverflow.com/questions/7358118/black-white-colormap-with-dashes-dots-etc
    """
    for ax in fig.get_axes():
        setAxLinesBW(ax)


def shp2vtk(df: GeoDataFrame, nodes=False) -> pv.PolyData:
    """
    Quickly convert a GeoDataFrame to a PolyData
    :param df: input GeoDataFrame
    :param nodes: If the geodataframe are points then set True
    :return: PolyData.

    Notes
    ------
    All the columns of the geodataframe, except for the geomtry column, will be written as cell data
    """

    get_coord_df = df.geometry.get_coordinates(ignore_index=False, index_parts=True)
    get_coord_df.reset_index(inplace=True)
    get_coord_df.columns = ['parts', 'indexes', 'x', 'y']
    get_coord_df['indexes'] = range(0, len(get_coord_df))
    get_coord_df['parts'] += 1

    get_coord_df['z'] = np.zeros(len(get_coord_df['x']))
    get_coord_df['points'] = get_coord_df.loc[:, ['x', 'y', 'z']].values.tolist()
    get_coord_df['points'] = get_coord_df['points'].map(tuple)
    duplicate_mask_first = get_coord_df.duplicated(subset=['x', 'y', 'z'], keep='last')
    first_repeats = get_coord_df[duplicate_mask_first]
    # other_repeats = get_coord_df[duplicate_mask_other]

    values = first_repeats['indexes'].values
    keys = first_repeats['points'].values

    unique_values_dict = dict(zip(keys, values))

    # to change the indexes as the ones that are in the dict we get the values from the dict with apply(get) and fill the na with the index column.
    get_coord_df['indexes'] = get_coord_df['points'].apply(func=lambda x: unique_values_dict.get(x)).fillna(
        get_coord_df['indexes']).astype(int)

    mem = 0
    conn = list()
    for part, index in zip(get_coord_df['parts'], get_coord_df.index):
        if part != mem:
            nparts = len(get_coord_df[get_coord_df['parts'] == part])
            conn.append(nparts)
        conn.append(get_coord_df.loc[index, 'indexes'])
        mem = part

    points = np.array(get_coord_df.loc[:, ['x', 'y', 'z']].values.tolist())
    if nodes:
        vtk_obj = pv.PolyData(points)
    else:
        vtk_obj = pv.PolyData(points, lines=conn)

    arrays = list(df.columns)
    arrays.remove('geometry')
    for array in arrays:
        vtk_obj.cell_data[array] = df[array].values


    # Use CleanPolyData to collapse double points in one

    p = 100000  # Scaling factor needed for the vtk function to work properly
    vtk_obj.points *= p
    clean = vtkCleanPolyData()
    clean.AddInputData(vtk_obj)
    clean.ToleranceIsAbsoluteOn()
    clean.ConvertLinesToPointsOff()
    clean.ConvertPolysToLinesOff()
    clean.ConvertStripsToPolysOff()
    clean.Update()

    output_obj = pv.PolyData(clean.GetOutput())
    output_obj.points /= p  # Rescale back the points

    return output_obj




"""

Collection of methods used to pass from a geopandas dataframe to a polydata or from
polydata to networkx. Each method is fit so that it plots nodes, fractures, boundary or the fracture network

TODO:
    + Add a adapter abstract class in the AbstractClasses.
    + Use multiblock instead of appender for fracture network (or all?)?
    + Try and find a way to crate boundary vtk object directly from polygon and not from lines of polygon

"""

import geopandas
import networkx
import numpy as np
from pyvista import PolyData, lines_from_points, MultiBlock
from shapely.geometry import MultiLineString

from vtkmodules.vtkFiltersCore import vtkAppendPolyData

import networkx as nx
from vtkmodules.vtkFiltersGeometry import vtkGeometryFilter

from fracability.operations.Geometry import connect_dots
from fracability.utils.general_use import shp2vtk


#  =============== VTK representations ===============

def node_vtk_rep(input_df: geopandas.GeoDataFrame) -> PolyData:
    points_vtk = shp2vtk(input_df)
    # points = np.array([point.coords for point in input_df.geometry]).reshape(-1, 3)
    # types = input_df['type'].values
    # node_types = input_df['n_type'].values
    #
    # points_vtk = PolyData(points)
    # points_vtk['type'] = types
    # points_vtk['n_type'] = node_types

    return points_vtk


def frac_vtk_rep(input_df: geopandas.GeoDataFrame) -> PolyData:

    conn_obj = shp2vtk(input_df)
    # appender = vtkAppendPolyData()
    #
    # for index, geom, set_n in zip(input_df.index, input_df['geometry'],
    #                               input_df['f_set']):  # For each geometry in the df
    #
    #     x, y = geom.coords.xy  # get xy as an array
    #     z = np.zeros_like(x)  # create a zeros z array with the same dim of the x (or y)
    #
    #     points = np.stack((x, y, z), axis=1)  # Stack the coordinates to a [n,3] shaped array
    #     # offset = np.round(points[0][0])
    #     if len(points) != 0:  # Avoid empty geometries
    #         pv_obj = lines_from_points(points)  # Create the corresponding vtk line with the given points
    #         pv_obj.cell_data['type'] = ['fracture'] * pv_obj.GetNumberOfCells()
    #         pv_obj.point_data['type'] = ['fracture'] * pv_obj.GetNumberOfPoints()
    #
    #         pv_obj.cell_data['f_set'] = [set_n] * pv_obj.GetNumberOfCells()
    #         pv_obj.point_data['f_set'] = [set_n] * pv_obj.GetNumberOfPoints()
    #
    #         pv_obj.cell_data['RegionId'] = [index] * pv_obj.GetNumberOfCells()
    #         pv_obj.point_data['RegionId'] = [index] * pv_obj.GetNumberOfPoints()
    #
    #         if 'length' in input_df.columns:
    #             pv_obj.cell_data['length'] = input_df.loc[index, 'length']
    #
    #     appender.AddInputData(pv_obj)  # Add the new object
    #
    # geometry_filter = vtkGeometryFilter()
    # geometry_filter.SetInputConnection(appender.GetOutputPort())
    # geometry_filter.Update()
    #
    # output_obj = PolyData(geometry_filter.GetOutput())
    # conn_obj = connect_dots(output_obj)

    return conn_obj


def bound_vtk_rep(input_df: geopandas.GeoDataFrame) -> PolyData:

    conn_obj = shp2vtk(input_df)
    # appender = vtkAppendPolyData()
    #
    # for index, geom, b_group in zip(input_df.index,
    #                                 input_df['geometry'],
    #                                 input_df['b_group']):  # For each geometry in the df
    #
    #     x, y = geom.coords.xy  # get xy as an array
    #     z = np.zeros_like(x)  # create a zeros z array with the same dim of the x (or y)
    #
    #     points = np.stack((x, y, z), axis=1)  # Stack the coordinates to a [n,3] shaped array
    #     # offset = np.round(points[0][0])
    #     pv_obj = lines_from_points(points)  # Create the corresponding vtk line with the given points
    #     pv_obj.cell_data['type'] = ['boundary'] * pv_obj.GetNumberOfCells()
    #     pv_obj.point_data['type'] = ['boundary'] * pv_obj.GetNumberOfPoints()
    #
    #     pv_obj.cell_data['b_group'] = [b_group] * pv_obj.GetNumberOfCells()
    #     pv_obj.point_data['b_group'] = [b_group] * pv_obj.GetNumberOfPoints()
    #
    #     pv_obj.cell_data['RegionId'] = [index] * pv_obj.GetNumberOfCells()
    #     pv_obj.point_data['RegionId'] = [index] * pv_obj.GetNumberOfPoints()
    #
    #
    #     # line.plot()
    #
    #     appender.AddInputData(pv_obj)  # Add the new object
    #
    # geometry_filter = vtkGeometryFilter()
    # geometry_filter.SetInputConnection(appender.GetOutputPort())
    # geometry_filter.Update()
    #
    # output_obj = PolyData(geometry_filter.GetOutput())
    # conn_obj = connect_dots(output_obj)

    return conn_obj


def fracture_network_vtk_rep(input_df: geopandas.GeoDataFrame, include_nodes=True) -> PolyData:

    fractures_df = input_df.loc[input_df['type'] == 'fracture']
    boundaries_df = input_df.loc[input_df['type'] == 'boundary']

    appender = vtkAppendPolyData()

    if include_nodes:
        nodes_df = input_df.loc[input_df['type'] == 'node']
        if not nodes_df.empty:
            nodes_vtk = node_vtk_rep(nodes_df)
            appender.AddInputData(nodes_vtk)

    if not fractures_df.empty:
        fractures_vtk = frac_vtk_rep(fractures_df)
        appender.AddInputData(fractures_vtk)
    if not boundaries_df.empty:
        boundaries_vtk = bound_vtk_rep(boundaries_df)
        appender.AddInputData(boundaries_vtk)

    appender.Update()

    geometry_filter = vtkGeometryFilter()
    geometry_filter.SetInputConnection(appender.GetOutputPort())
    geometry_filter.Update()

    output_obj = PolyData(geometry_filter.GetOutput())
    conn_obj = connect_dots(output_obj)

    return conn_obj


#  =============== Networkx representations ===============


def networkx_rep(input_object: PolyData) -> networkx.Graph():

    network = input_object
    lines = network.lines  # Get the connectivity list of the object

    lines = np.delete(lines,
                      np.arange(0, lines.size, 3)).reshape(-1, 2)  # remove padding eg. [2 id1 id2 2 id3 id4 ...] -> remove the 2

    network = nx.Graph()  # Create a networkx graph instance

    network.add_edges_from(lines)  # Add the edges

    output_obj = network
    return output_obj





import numpy as np
import shapely.geometry as geom
from shapely.affinity import scale
from shapely.ops import split

def int_node(line1, line2, idx_list, gdf):
    """
    Function used to add the intersection node to a line crossed or touched by a second line.
    This function works by extending of a given factor the start and end segment the line2 and then use shapely to:

    1. Split line1 with line2
    2. Join the two split geometries

    We do this to identify and add the intersection node to line1:

    O----------O --split--> O-----O O-----O --join--> O-----O-----O

    The extension is used in case the joint is a Y or T and the positioning of the node is not pixel perfect. Depending
    on the number of segment composing line2 we extend in different ways:

    1. Only one segment extend the whole thing using the start and end vertex
    2. Two segments extend the two segments using the end of the first and the start of the second.
    3. Three or more segments extend the first and last segment using the end of the first and the start of the last.
    """
    new_geom_dict = {}
    fac = 1.05

    # line1_points = np.array(line1.coords)
    # line2_points = np.array(line2.coords)
    #
    # unique_points1 = np.unique(line1_points, axis=0)
    # unique_points2 = np.unique(line2_points, axis=0)
    #
    # line1 = geom.LineString(unique_points1)
    # line2 = geom.LineString(unique_points2)

    try:
        if line1.crosses(line2):
            split_lines1 = split(line1, line2)
            outcoords1 = [list(i.coords) for i in split_lines1.geoms]
            split_lines2 = split(line2, line1)
            outcoords2 = [list(i.coords) for i in split_lines2.geoms]

            new_line1 = geom.LineString([i for sublist in outcoords1 for i in sublist])
            new_line2 = geom.LineString([i for sublist in outcoords2 for i in sublist])
            new_geom_dict[idx_list[0]] = new_line1
            new_geom_dict[idx_list[1]] = new_line2
        else:
            for counter in range(3):
                if len(line2.coords) == 2:
                    scaled_segment1 = scale(line2, xfact=fac, yfact=fac, origin=line2.boundary.geoms[0])
                    scaled_segment2 = scale(scaled_segment1, xfact=fac, yfact=fac, origin=scaled_segment1.boundary.geoms[1])
                    extended_line = geom.LineString(scaled_segment2)
                elif len(line2.coords) == 3:
                    first_seg = geom.LineString(line2.coords[:2])
                    last_seg = geom.LineString(line2.coords[-2:])
                    scaled_first_segment = scale(first_seg, xfact=fac, yfact=fac, origin=first_seg.boundary.geoms[1])
                    scaled_last_segment = scale(last_seg, xfact=fac, yfact=fac, origin=last_seg.boundary.geoms[0])
                    extended_line = geom.LineString([*scaled_first_segment.coords, *scaled_last_segment.coords])
                else:
                    first_seg = geom.LineString(line2.coords[:2])
                    last_seg = geom.LineString(line2.coords[-2:])
                    # print(np.array(first_seg.boundary.geoms), idx_list)
                    scaled_first_segment = scale(first_seg, xfact=fac, yfact=fac, origin=first_seg.boundary.geoms[1])
                    scaled_last_segment = scale(last_seg, xfact=fac, yfact=fac, origin=last_seg.boundary.geoms[0])
                    extended_line = geom.LineString(
                        [*scaled_first_segment.coords, *line2.coords[2:-2], *scaled_last_segment.coords])

                split_lines = split(line1, extended_line)

                # Check if the selected line combination defined a intersection. If not, switch the lines.
                # If again no intersection is found then there is no intersection
                # If found then create a LineString composed of the two segments (we don't use merge because this is the
                # only way to be sure that no multipart object is created; merge creates multipart if the two points are not
                # exactly overlapping).

                if len(split_lines.geoms) == 1 and counter < 2:
                    line1, line2 = line2, line1
                elif len(split_lines.geoms) == 1 and counter >= 2:
                    new_geom_dict = {}
                else:
                    outcoords = [list(i.coords) for i in split_lines.geoms]
                    new_line = geom.LineString([i for sublist in outcoords for i in sublist])
                    new_geom_dict[idx_list[counter]] = new_line

                    # # Plot results to visualize intersections

                    # test2 = np.array(extended_line.coords)
                    # test1 = np.array(new_line.coords)
                    #
                    # plt.plot(test1[:, 0], test1[:, 1], 'r-o')
                    # plt.plot(test2[:, 0], test2[:, 1], 'b-o')
                    # plt.show()

                    break
    except IndexError:
        print(f'Possible duplicate point found, fix geometry on gis:')
        obj1_idx = gdf.loc[idx_list[0], 'og_line_id']
        set1_idx = gdf.loc[idx_list[0], 'f_set']
        obj2_idx = gdf.loc[idx_list[1], 'og_line_id']
        set2_idx = gdf.loc[idx_list[1], 'f_set']

        if set2_idx == -9999:
            group2_idx = gdf.loc[idx_list[1], 'b_group']
            print(f'lines {obj1_idx} (set {set1_idx}), {obj2_idx} (boundary group {group2_idx})')
        elif set1_idx == -9999:
            group1_idx = gdf.loc[idx_list[0], 'b_group']
            print(f'lines {obj1_idx} (boundary group {group1_idx}), {obj2_idx} (set {set2_idx})')
        else:
            print(f'lines {obj1_idx} (set {set1_idx}), {obj2_idx} (set {set2_idx})')
        exit()
    except ValueError:
        obj1_idx = gdf.loc[idx_list[0], 'og_line_id']
        set1_idx = gdf.loc[idx_list[0], 'f_set']
        obj2_idx = gdf.loc[idx_list[1], 'og_line_id']
        set2_idx = gdf.loc[idx_list[1], 'f_set']

        if set2_idx == -9999:
            group2_idx = gdf.loc[idx_list[1], 'b_group']
            print(f'lines {obj1_idx} (set {set1_idx}), {obj2_idx} (boundary group {group2_idx}) overlapping, checking if single or multiple point overlap')
        elif set1_idx == -9999:
            group1_idx = gdf.loc[idx_list[0], 'b_group']
            print(f'lines {obj1_idx} (boundary group {group1_idx}), {obj2_idx} (set {set2_idx}) overlapping, checking if single or multiple point overlap')
        else:
            print(f'lines {obj1_idx} (set {set1_idx}), {obj2_idx} (set {set2_idx}) overlapping, checking if single or multiple point overlap')

        element = np.array(line1.coords)
        test_element = np.array(extended_line.coords)
        mask = np.isin(element, test_element)

        if len(element[mask]) == 4:
            print('Single point overlap. Geometry is correct, continuing. Shp files could contain invalid geometries, run the check_network method to perform a network wide check.')
            pass
        else:
            print('Multiple point overlap. Check and fix geometry, stopping. Shp files could contain invalid geometries, run the check_network method to perform a network wide check.')
            exit()

        # print(gdf.loc[idx_list[0]-5:idx_list[1]+5])
    return new_geom_dict





import os.path
import shutil
from abc import ABC, abstractmethod, abstractproperty
from fracability.examples.data import QgisStyle


from geopandas import GeoDataFrame
from geopandas import read_file
from pyvista import PolyData
from networkx import Graph
import scipy.stats as ss
import numpy as np
from copy import deepcopy
from shapely import remove_repeated_points


class BaseEntity(ABC):
    """
    Abstract class for Fracture network entities:

    1. Nodes
    2. Fractures
    3. Boundaries
    4. Fracture Networks
    """

    def __init__(self, gdf: GeoDataFrame = None, csv: str = None, shp: str = None):
        """
        Init the entity. If a geopandas dataframe is specified then it is
        set as the source entity df.

        Parameters
        -----------
        gdf: GeoDataFrame
            Use as input a geopandas dataframe
        csv: str
            Use as input a csv indicated by the path
        shp: str
            Use as input a shapefile indicated by the path

        Notes
        --------
        The csv needs to have a "geometry" column. If missing the import will fail.
        """

        self._df: GeoDataFrame = GeoDataFrame()
        if gdf is not None:
            self.entity_df = gdf
        elif csv is not None:
            gdf = read_file(csv, GEOM_POSSIBLE_NAMES="geometry", KEEP_GEOM_COLUMNS="NO")
            if 'geometry' not in self.entity_df.columns:
                exit('Missing geometry column, terminating')
            else:
                self.entity_df = gdf
        elif shp is not None:
            self.entity_df = read_file(shp)

    @property
    def name(self) -> str:
        """
        Property used to return the name of the current class as a string.
        """
        return self.__class__.__name__

    @property
    @abstractmethod
    def entity_df(self) -> GeoDataFrame:
        """
        Each entity is based on a geopandas dataframe. This property returns or sets
        the entity_df of the given entity.

        :getter: Returns the GeoDataFrame
        :setter: Sets the GeoDataFrame
        :type: GeoDataFrame

        Notes
        -------
        When set, the dataframe is modified to conform to the assigned entity structure.
        """

        pass

    @entity_df.setter
    def entity_df(self, gpd: GeoDataFrame = None):
        self._df = gpd

    @property
    @abstractmethod
    def vtk_object(self) -> PolyData:
        """
        Each entity can be represented with a vtk object.
        This returns a Pyvista PolyData object representing the entity_df.

        :getter: Returns a Pyvista PolyData object
        :setter: Sets a generic Pyvista DataSet
        :type: pyvista PolyData

        Notes
        -------
        When the get method is applied the PolyData is build **on the fly** using the entity_df as a source.

        When set the DataSet is **cast to a PolyData**.
        """
        pass

    @property
    @abstractmethod
    def network_object(self) -> Graph:
        """
        Each entity can be represented with a networkx graph.
        This returns the network object using the vtk object (and so the df).

        :getter: Returns a networkx Graph object
        :setter: Sets a Graph object
        :type: pyvista Graph

        Notes
        -------
        When the get method is applied the Graph is build **on the fly** using the object and so the entity_df.
        """

        pass

    @abstractmethod
    def mat_plot(self):
        """
        Plot entity using matplotlib backend
        """

    @abstractmethod
    def vtk_plot(self):
        """
        Plot entity using vtk backend
        """

    @property
    def crs(self) -> str:
        """
        Property used to return the crs of the entity

        :return: Name of the coordinate system as a string
        """
        return self.entity_df.crs

    @crs.setter
    def crs(self, crs):
        """
        Property used to return the crs of the entity

        :return: Name of the coordinate system as a string
        """
        self.entity_df.crs = crs

    @property
    def centroid(self) -> np.ndarray:
        """
        Property used to return the centroid of the entity. Dissolve is used to aggregate each shape in a single entity.

        :return: 1D numpy array of the centroid
        """
        trans_center = np.array(self.entity_df.dissolve().centroid[0].coords).flatten()
        return trans_center

    @property
    def get_copy(self):
        """
        Property used to return a deep copy of the entity

        :return:
        """
        return deepcopy(self)

    def center_object(self, trans_center: np.array = None, return_center: bool = False,
                      inplace: bool = True):
        """
        Method used to center the center of an Entity object to a given point. If no point is specified then the object
        will be moved to the origin (0,0,0).

        Parameters
        ----------
        obj: Boundary, Fractures, FractureNetwork
            A fracability entity object
        trans_center: array
            Point to which translate the object
        return_center: bool
            Bool flag to specify to return the translation center
        inplace: bool
            Bool flag to specify if the operation overwrites the entity or creates a new instance

        Returns
        ----------
        trans_center: array
            Point of translation. If trans_center is not specified in the output then this will return the center of
            the object
        copy_obj: object
            Copy of the modified input object (preserves the original input)
        """

        if self.name == 'FractureNetwork':
            df = self.fracture_network_to_components_df()
        else:
            df = self.entity_df.copy()

        if trans_center is None:
            trans_center = self.centroid

        df['geometry'] = df.translate(-trans_center[0], -trans_center[1])

        if inplace:

            self.entity_df = df

            if return_center:
                return trans_center

        else:
            copy_obj = deepcopy(self)
            copy_obj.entity_df = df

            if return_center:
                return copy_obj, trans_center
            else:
                return copy_obj

    def save_csv(self, path: str, sep: str = ';', index: bool = False):
        """
        Save the entity df as csv

        Parameters
        -------------
        index: Bool
            Indicate whether to include or not the index column
        sep: String
            Indicate the separator string used to save the csv
        path: String
            Indicate the path in where to save the csv. **DO NOT** include the extension (.csv).

        Notes
        ---------

        The csv will be saved in the output/csv directory in the indicated path in the working directory. If this path does not exist, then it will be created.
        """

        if not self.entity_df.empty:
            cwd = os.getcwd()
            output_path = os.path.join(cwd, path, 'output/csv')

            if not os.path.isdir(output_path):
                os.makedirs(output_path)

            if self.name == 'Fractures':
                set_n = list(set(self.entity_df['f_set']))
                for f_set in set_n:
                    final_path = os.path.join(output_path, f'{self.name}_{f_set}.csv')
                    entity_df = self.entity_df.loc[self.entity_df['f_set'] == f_set, :]
                    entity_df.to_csv(final_path, sep=sep, index=index)
            elif self.name == 'Boundary':
                group_n = list(set(self.entity_df['b_group']))
                for b_group in group_n:
                    final_path = os.path.join(output_path, f'{self.name}_{b_group}.csv')
                    entity_df = self.entity_df.loc[self.entity_df['b_group'] == b_group, :]
                    entity_df.to_csv(final_path, sep=sep, index=index)
            else:
                final_path = os.path.join(output_path, f'{self.name}.csv')
                entity_df = self.entity_df
                entity_df.to_csv(final_path, sep=sep, index=index)



        else:
            print('Cannot save an empty entity')

    def save_shp(self, path: str):
        """
        Save the entity df as shapefile

        Parameters
        -------------
        path: String.
            Indicate the path in where to save the csv. **DO NOT** include the extension (.shp).

        Notes
        ---------

        The shapefile will be saved in output/shp directory in the indicated path in the working directory. If this path does not exist, then it will be created.
        """
        if not self.entity_df.empty:
            cwd = os.getcwd()
            output_path = os.path.join(cwd, path, 'output', 'shp')
            style_out = os.path.join(output_path, 'qgis_style')
            if not os.path.isdir(output_path):
                os.makedirs(output_path)
            if not os.path.isdir(style_out):
                os.makedirs(style_out)

            if self.name == 'Fractures' or self.name == 'Backbone':
                set_n = list(set(self.entity_df['f_set']))
                for f_set in set_n:
                    final_path = os.path.join(output_path, f'{self.name}_{f_set}.shp')
                    entity_df = self.entity_df.loc[self.entity_df['f_set'] == f_set, :]
                    entity_df.to_file(final_path, crs=self.crs)
            elif self.name == 'Boundary':
                group_n = list(set(self.entity_df['b_group']))
                for b_group in group_n:
                    final_path = os.path.join(output_path, f'{self.name}_{b_group}.shp')
                    entity_df = self.entity_df.loc[self.entity_df['b_group'] == b_group, :]
                    entity_df.to_file(final_path, crs=self.crs)
            elif self.name == 'Nodes':
                final_path = os.path.join(output_path, f'{self.name}.shp')
                entity_df = self.entity_df
                entity_df.to_file(final_path, crs=self.crs)

            qgis_style_paths = QgisStyle().available_paths

            for qgis_path in qgis_style_paths:
                shutil.copy(qgis_path, style_out)

        else:
            print('Cannot save an empty entity')

    def remove_double_points(self):
        """
        Utility used to clean geometries with double points
        """
        tot_geom = len(self.entity_df.geometry)
        print('\n\n')
        for line, geom in enumerate(self.entity_df.geometry):
            print(f'Removing possible double points on geometries: {line}/{tot_geom}',end='\r')
            self.entity_df.loc[line, 'geometry'] = remove_repeated_points(geom, tolerance=0.000001)


class BaseOperator(ABC):
    """
    Abstract class for Operators such as:

    1. Geometry operations
    2. Topology operations

    This class provides a unified input for the different operators since all are based on the BaseObj and associated
    entity_df
    """

    def __init__(self, obj: BaseEntity):
        self.df = obj.entity_df.loc[obj.entity_df['active_set'] == 1].copy()
        self.obj = obj

    @property
    def name(self):
        return self.__class__.__name__


class AbstractStatistics(ABC):

    def __init__(self, obj: BaseEntity = None):
        self._obj: BaseEntity = obj

        self.data: ss.CensoredData
        self._lengths: np.array

        self._function_list: list = ['pdf', 'cdf', 'sf', 'hf', 'chf']

        if obj is not None:
            self._lengths = obj.entity_df.loc[obj.entity_df['censored'] >= 0, 'length'].values
            entity_df = self._obj.entity_df
            self._complete_lengths = entity_df.loc[entity_df['censored'] == 0, 'length'].values
            self._censored_lengths = entity_df.loc[entity_df['censored'] == 1, 'length'].values

            self.data = ss.CensoredData(uncensored=self._complete_lengths, right=self._censored_lengths)

    @property
    def lengths(self):

        """
        This property returns or sets the complete list of length data for the fracture network

        :getter: Return the complete list of lengths
        :setter: Set the complete list of lengths
        """

        return self._lengths

    @lengths.setter
    def lengths(self, length_list: list = None):
        self._lengths = length_list

    @property
    def complete_lengths(self):

        """
        This property returns or sets the list of non-censored length data of the fracture network

        :getter: Return the list of non-censored data
        :setter: Set the list of non-censored data
        """

        return self.data.__dict__['_uncensored']

    @complete_lengths.setter
    def complete_lengths(self, complete_length_list: list = []):
        self.data = ss.CensoredData(uncensored=complete_length_list, right=self.censored_lengths)

    @property
    def censored_lengths(self):

        """
        This property returns or sets the list of censored length data of the fracture network

        :getter: Return the list of censored data
        :setter: Set the list of censored data
        :return:
        """

        return self.data.__dict__['_right']

    @censored_lengths.setter
    def censored_lengths(self, censored_length_list: list = []):
        self.data = ss.CensoredData(uncensored=self.complete_lengths, right=censored_length_list)

    @property
    def ecdf(self):
        return ss.ecdf(self.data)

    @property
    def function_list(self):
        return self._function_list




from copy import deepcopy

from pyvista import PolyData
from vtkmodules.vtkFiltersCore import vtkCleanPolyData

from fracability.AbstractClasses import BaseEntity
from fracability.utils.shp_operations import int_node

def connect_dots(vtk_obj: PolyData) -> PolyData:

    """Method used to clean intersection in vtk objects (for example merge two overlapping nodes)"""

    p = 100000  # Scaling factor needed for the vtk function to work properly
    vtk_obj.points *= p
    clean = vtkCleanPolyData()
    clean.AddInputData(vtk_obj)
    clean.ToleranceIsAbsoluteOn()
    clean.ConvertLinesToPointsOff()
    clean.ConvertPolysToLinesOff()
    clean.ConvertStripsToPolysOff()
    clean.Update()

    output_obj = PolyData(clean.GetOutput())
    output_obj.points /= p  # Rescale back the points

    return output_obj


#@Halo(text='Calculating intersections', spinner='line', placement='right')
def tidy_intersections(obj, buffer=0.05, inplace: bool = True):
    """Method used to tidy shapefile intersections between fractures in a fracture or fracture network object."""

    if obj.name == 'FractureNetwork':
        gdf = obj.fracture_network_to_components_df()
        gdf = gdf.loc[gdf['type'] != 'node']
    elif obj.name == 'Fractures':
        gdf = obj.entity_df.copy()
    else:
        print('Cannot tidy intersection for nodes or only boundaries')
        return

    df_buffer = gdf.buffer(buffer)
    print('\n\n')
    for idx_line1, line in gdf.iterrows():
        print(f'Calculating intersections on fracture: {idx_line1+1}/{len(gdf.index)}', end='\r')

        if line['type'] == 'boundary':
            continue

        line1 = line['geometry']

        idx_list = df_buffer.index[df_buffer.intersects(line1) == True]  # Subset the intersecting lines
        idx_list = idx_list[idx_list != idx_line1]  # Exclude the reference line index to avoid self intersection

        intersections = gdf.loc[idx_list]
        for line2, idx_line2 in zip(intersections['geometry'], idx_list):  # for each intersecting line:
            new_geom = int_node(line1, line2, [idx_line1, idx_line2], gdf)  # Calculate and add the intersection node.

            for key, value in new_geom.items():
                gdf.loc[key, 'geometry'] = value  # substitute the original geometry with the new geometry

            line1 = gdf.loc[
                idx_line1, 'geometry']  # Use as the reference line (in the int_node function) the new geometry.

    if inplace:
        obj.entity_df = gdf
    else:
        copy_obj = deepcopy(obj)
        copy_obj.entity_df = gdf
        return copy_obj


def calculate_seg_length(obj: BaseEntity, inplace: bool = True):
    """Method used to calculate and set fracture lengths when absent"""

    if obj.name == 'FractureNetwork':
        gdf = obj.fracture_network_to_components_df()
        gdf = gdf.loc[gdf['type'] != 'node']
        gdf = gdf.loc[gdf['type'] != 'boundary']
    elif obj.name == 'Nodes':
        print('cannot calculate lengths of nodes')
        return
    else:
        gdf = obj.entity_df.copy()

    df = gdf.copy()

    if 'lengths' in df.columns:
        print('length column already present. Delete it to recalculate lengths')
    else:
        df['lengths'] = df.length

    if inplace:
        obj.entity_df = df
    else:
        copy_obj = deepcopy(obj)
        copy_obj.entity_df = df
        return copy_obj




import numpy as np

from numpy import exp
from numpy import log as ln
import pandas as pd
from pandas import DataFrame
import scipy.stats as ss
from scipy.optimize import minimize

from fracability.utils.general_use import KM
import fracability.Plotters as plotter


class NetworkData:

    """ Class used to represent fracture or fracture network data.
    It acts as a wrapper for the scipy CensoredData class

    :param obj: fracture/fracture network object
    :param use_survival: Use survival analysis to get the distribution. If false the whole dataset is used considering all fractures complete
    :param complete_only: When not using survival, use only the complete length values
    """

    def __init__(self, obj=None, use_survival=True, complete_only=True):


        self._obj = obj  # fracture/fracture network object

        self._data = None  # the data
        self._lengths: np.ndarray  # array of all the lengths (both complete and censored)
        self._non_censored_lengths: np.ndarray  # array of only complete lengths
        self._censored_lengths: np.ndarray  # array of only censored lengths
        self._ecdf: np.ndarray
        self._function_list: list = ['pdf', 'cdf', 'sf', 'hf', 'chf']  # list of possible functions

        self.use_survival = use_survival
        self.complete_only = complete_only

        if obj is not None:
            if self._obj.name == 'FractureNetwork':
                frac_obj = self._obj.fractures
            else:
                frac_obj = self._obj

            entity_df = frac_obj.entity_df
            entity_df = entity_df.sort_values(by='length')

            self.delta = 1-entity_df['censored'].values
            self._lengths = entity_df['length'].values
            self._non_censored_lengths = entity_df.loc[entity_df['censored'] == 0, 'length'].values
            self._censored_lengths = entity_df.loc[entity_df['censored'] == 1, 'length'].values
            self._ecdf = KM(self._lengths, self._lengths, self.delta)

            if self.use_survival:
                self.data = ss.CensoredData(uncensored=self._non_censored_lengths, right=self._censored_lengths)
            else:
                if self.complete_only:
                    self.data = self.non_censored_lengths
                else:
                    self.delta[:] = 1
                    self.data = self.lengths

    @property
    def data(self):
        """
        Property that returns or sets the CensoredData class of the fracture network
        :return:
        """
        return self._data

    @data.setter
    def data(self, data):
        self._data = data

    @property
    def lengths(self) -> np.ndarray:

        """
        This property returns or sets the complete list of length data (censored and uncesored) for the fracture network

        :getter: Return the complete list of lengths
        :setter: Set the complete list of lengths
        """

        return self._lengths

    @property
    def non_censored_lengths(self) -> np.ndarray:

        """
        This property returns or sets the list of non-censored length data of the fracture network

        :getter: Return the list of non-censored data
        :setter: Set the list of non-censored data
        """

        return self._non_censored_lengths

    @property
    def censored_lengths(self) -> np.ndarray:

        """
        This property returns or sets the list of censored length data of the fracture network

        :getter: Return the list of censored data
        :setter: Set the list of censored data
        :return:
        """

        return self._censored_lengths

    @property
    def function_list(self) -> list:
        """
        Property that returns the list of available probability functions (pdf, cdf etc)
        :return:
        """
        return self._function_list

    @property
    def mean(self) -> np.ndarray:
        """
        Calculate the sample mean of ALL the input data (i.e. it ignores the complete_only flag).
        :return: Numpy array of the sample mean
        """
        return np.mean(self.lengths)

    @property
    def std(self) -> np.ndarray:
        """
        Calculate the sample standard deviation of ALL the input data (i.e. it ignores the complete_only flag).
        :return: Numpy array of the sample standard deviation
        """
        return np.std(self.lengths)

    @property
    def var(self) -> np.ndarray:
        """
        Calculate the sample variance of ALL the input data (i.e. it ignores the complete_only flag).
        :return: Numpy array of the sample variance
        """
        return np.var(self.lengths)

    @property
    def median(self) -> np.ndarray:
        """
        Calculate the sample median of ALL the input data (i.e. it ignores the complete_only flag).
        :return: Numpy array of the sample median
        """
        return np.median(self.lengths)

    @property
    def mode(self) -> tuple:
        """
        Calculate the sample mode of ALL the input data (i.e. it ignores the complete_only flag).
        :return: Tuple of the most frequent value(s)
        """
        return ss.mode(self.lengths)

    @property
    def b5(self) -> np.ndarray:
        """
        Calculate the sample 5th percentile of ALL the input data (i.e. it ignores the complete_only flag).
        :return: Numpy array of the sample 5th percentile
        """
        return np.percentile(self.lengths, 5)

    @property
    def b95(self) -> np.ndarray:
        """
        Calculate the sample 95th percentile of ALL the input data (i.e. it ignores the complete_only flag).
        :return: Numpy array of the sample 95th percentile
        """
        return np.percentile(self.lengths, 95)

    @property
    def ecdf(self) -> np.ndarray:
        """
        Property that returns the empirical CDF of the input data (it ignores the complete_only flag but ) using Kaplan-Meier.
        :return: Numpy array of the calculated CDF values using KM
        """
        return self._ecdf

    @property
    def esf(self) -> np.ndarray:
        """
        Property that returns the empirical SF of the ALL the input data (i.e. it ignores the complete_only flag) using Kaplan-Meier.
        :return: Numpy array of the calculated SF values using KM
        """
        return 1-self.ecdf

    @property
    def total_n_fractures(self) -> int:
        """
        Total number of fractures of ALl the input data (i.e. it ignores the complete_only flag)
        :return: Int. Number of fractures
        """
        return len(self.lengths)

    @property
    def censoring_percentage(self) -> float:
        """
        Percentage of censoring of ALl the input data (i.e. it ignores the complete_only flag)
        :return: Float. Censoring %
        """
        return (len(self.censored_lengths)/self.total_n_fractures)*100


class NetworkDistribution:
    """
    Class used to represent a fracture or fracture network length distribution. It is essentially a wrapper for the
    scipy rv distributions.
    """

    def __init__(self, parent, obj: ss.rv_continuous = None, parameters: tuple = None, fit_data: NetworkData = None):

        self.parent: NetworkFitter = parent
        self._distribution = obj.freeze(*parameters)
        self.fit_data = fit_data

    @property
    def distribution(self) -> ss.rv_continuous:
        """
        Property that returns or sets a *frozen* ss.rv_continuous class
        :return:
        """
        return self._distribution

    @distribution.setter
    def distribution(self, distribution: ss.rv_continuous = None):
        self._distribution = distribution

    @property
    def distribution_name(self) -> str:
        """
        Property that returns the name of the given distribution
        :return:
        """
        return self.distribution.dist.name

    @property
    def distribution_parameters(self) -> tuple:
        """
        Property that returns the parameters of the frozen distribution
        :return:
        """
        return self.distribution.args

    @property
    def n_distribution_parameters(self) -> int:
        """
        Property that returns the number of parameters of the frozen distribution. the args method returns the
        shape parameters. This means that except for normal and logistic the loc counts as a parameter. To fix this we
        subtract -1 to all except normal and logistic.
        :return:
        """
        if self.distribution_name == 'norm' or self.distribution_name == 'logistic':

            return len(self.distribution_parameters)
        else:
            return len(self.distribution_parameters)-1

    @property
    def mean(self) -> float:
        """
        Property that returns the mean of the frozen distribution
        :return:
        """
        return self.distribution.mean()

    @property
    def mode(self) -> list:
        """
        Property that returns the mode(s) of the pdf of the given distribution
        :return:
        """
        return minimize(lambda x: -self.distribution.pdf(x), np.ceil(self.distribution_parameters[-1])).x

    @property
    def median(self) -> float:
        """
        Property that returns the median of the distribution
        :return:
        """
        return self.distribution.median()

    @property
    def var(self) -> float:
        """
        Property that returns the variance of the distribution
        :return:
        """
        return self.distribution.var()

    @property
    def std(self) -> float:
        """
        Property that returns the standard deviation of the distribution
        :return:
        """
        return self.distribution.std()

    @property
    def b5(self) -> float:
        """
        Property that returns the 5th percentile of the distribution
        :return:
        """
        return self.distribution.ppf(0.05)

    @property
    def b95(self) -> float:
        """
        Property that returns the 95th percentile of the distribution
        :return:
        """
        return self.distribution.ppf(0.95)

    def cdf(self, x_values: np.array = None):
        if x_values is not None:
            return self.distribution.cdf(x_values)
        else:
            return self.distribution.cdf(self.fit_data.lengths)

    def log_pdf(self, x_values: np.array = None) -> np.array:
        """
        Property that returns the logpdf calculated on the data
        :return:
        """

        if x_values.any():
            return self.distribution.logpdf(x_values)
        else:
            return np.array([0])

    def log_sf(self, x_values: np.array = None) -> np.array:
        """
        Property that returns the logsf calculated on the data
        :return:
        """
        if x_values.any():
            return self.distribution.logsf(x_values)
        else:
            return np.array([0])

    @property
    def max_log_likelihood(self) -> float:
        """
        Property that returns the log likelihood of the distribution. The likelihood is calculated by adding
        the cumulative sum of the log pdf and log sf of the fitted distribution.
        :return:
        """
        if self.fit_data.use_survival:
            log_f = self.log_pdf(self.fit_data.non_censored_lengths)
            log_r = self.log_sf(self.fit_data.censored_lengths)
        else:
            if self.fit_data.complete_only:
                sample = self.fit_data.non_censored_lengths
            else:
                sample = self.fit_data.lengths
            log_f = self.log_pdf(sample)
            log_r = np.array([0])

        LL_f = log_f.sum()
        LL_rc = log_r.sum()

        return LL_f + LL_rc

    # Distance parameters

    @property
    def AIC(self) -> float:
        """
        Property that returns the classic Akaike Information Criterion (1974) of the distribution
        :return:
        """
        max_log_likelihood = self.max_log_likelihood

        k = self.n_distribution_parameters

        AIC = (-2 * max_log_likelihood) + (2 * k)

        return AIC

    @property
    def AICc(self) -> float:
        """
        Property that returns the Akaike Information Criterion (for small number of values) of the distribution
        :return:
        """

        LL2 = -2 * self.max_log_likelihood

        k = self.n_distribution_parameters
        n = self.fit_data.total_n_fractures

        if (n - k - 1) == 0: # if n parameters + 1 = n of total fractures then the akaike is invalid i.e. the total fractures must be > than the n_parameters+1
            return -1
        else:
            AICc = 2 * k + LL2 + (2 * k ** 2 + 2 * k) / (n - k - 1)
            return AICc

    @property
    def Akaike_rank(self):
        fitter_records = self.parent.fit_records()
        name = self.distribution_name
        return fitter_records.loc[fitter_records['name'] == name, 'Akaike_rank'].values[0]

    @property
    def BIC(self) -> float:
        """
        Property that returns the Bayesian Information Criterion of the distribution
        :return:
        """
        LL2 = -2 * self.max_log_likelihood

        k = len(self.distribution_parameters)
        n = len(self.fit_data.non_censored_lengths) + len(self.fit_data.censored_lengths)

        BIC = np.log(n)*k + LL2
        return BIC

    @property
    def KS_distance(self) -> float:
        """
        Calcuate the Kolmogorov-Smirnov distance between the empirical and the fitted model
        :return: Float. The KS distance

        Notes
        ------
        Kim 2019, Tests based on EDF statistics for randomly censored normal
        distributions when parameters are unknown
        """
        Z = self.distribution.cdf(self.fit_data.lengths)
        G_n = self.fit_data.ecdf
        tot_n = self.fit_data.total_n_fractures
        delta = self.fit_data.delta

        complete_list_index = np.where(delta == 1)[0]
        diff_plus_list = []
        diff_minus_list = []

        for j in complete_list_index:
            diff_plus = G_n[j] - Z[j]  # Calculate the positive difference at index j (DC+)

            if j + 1 > tot_n - 1:  # Check if j+1 (0 indexed) is bigger than the total number tot_n (1 indexed)
                Z_j1 = 1
            else:
                Z_j1 = Z[j + 1]

            diff_minus = Z_j1 - G_n[j]  # Calculate the negative difference at index j (DC-)
            diff_plus_list.append(diff_plus)
            diff_minus_list.append(diff_minus)

        DCn_pos = max(diff_plus_list)
        DCn_neg = max(diff_minus_list)

        DCn = max(DCn_pos, DCn_neg)

        return DCn

    @property
    def KS_rank(self):
        fitter_records = self.parent.fit_records()
        name = self.distribution_name
        return fitter_records.loc[fitter_records['name'] == name, 'KS_rank'].values[0]

    @property
    def KG_distance(self) -> float:
        """
        Calcuate the Koziol and Green distance between the empirical and the fitted model
        :return: Float. The KG distance

        Notes
        ------
        Kim 2019, Tests based on EDF statistics for randomly censored normal
        distributions when parameters are unknown
        """

        Z = self.distribution.cdf(self.fit_data.lengths)
        G_n = self.fit_data.ecdf
        tot_n = self.fit_data.total_n_fractures
        kg_sum = 0

        for j in range(tot_n):

            if j + 1 > tot_n - 1:  # Check if j+1 (0 indexed) is bigger than the total number tot_n (1 indexed)
                Z_j1 = 1
            else:
                Z_j1 = Z[j + 1]
            kg_sum += G_n[j] * (Z_j1 - Z[j]) * (G_n[j] - (Z_j1 + Z[j]))

        psi_sq = (tot_n * kg_sum) + tot_n / 3

        return psi_sq

    @property
    def KG_rank(self):
        fitter_records = self.parent.fit_records()
        name = self.distribution_name
        return fitter_records.loc[fitter_records['name'] == name, 'KG_rank'].values[0]

    @property
    def AD_distance(self) -> float:
        """
        Calcuate the Koziol and Green distance between the empirical and the fitted model
        :return: Float. The KG distance

        Notes
        ------
        Kim 2019, Tests based on EDF statistics for randomly censored normal
        distributions when parameters are unknown
        """
        smallest_number = 10**-10
        Z = self.distribution.cdf(self.fit_data.lengths)
        G_n = self.fit_data.ecdf
        tot_n = self.fit_data.total_n_fractures

        sum1 = 0  # First sum
        sum2 = 0  # Second sum
        if Z[-1] == 1:
            Z[-1] -= smallest_number  # this is to avoid 0 in ln(1 - Z[-1])
        elif Z[-1] == 0:
            Z[-1] += smallest_number  # this is to avoid 0 in ln(Z[-1])
        for j in range(tot_n - 1):

            if Z[j + 1] == 0:
                Z[j + 1] += smallest_number  # this is to avoid 0 in ln(Z[j + 1])
            elif Z[j + 1] == 1:
                Z[j + 1] -= smallest_number  # this is to avoid 0 in ln(1 - Z[j + 1])

            if Z[j] == 0:
                Z[j] += smallest_number  # this is to avoid 0 in ln(Z[j])
            elif Z[j] == 1:
                Z[j] -= smallest_number  # this is to avoid 0 in ln(1 - Z[j])

            sum1 += (G_n[j] ** 2) * (-ln(1 - Z[j + 1]) + ln(Z[j + 1]) + ln(1 - Z[j]) - ln(Z[j]))
            sum2 += G_n[j] * (-ln(1 - Z[j + 1]) + ln(1 - Z[j]))

        AC_sq = (tot_n * sum1) - (2 * tot_n * sum2) - (tot_n * ln(1 - Z[-1])) - (tot_n * ln(Z[-1])) - tot_n

        return AC_sq

    @property
    def AD_rank(self):
        fitter_records = self.parent.fit_records()
        name = self.distribution_name
        return fitter_records.loc[fitter_records['name'] == name, 'AD_rank'].values[0]

    @property
    def Mean_rank(self):
        fitter_records = self.parent.fit_records()
        name = self.distribution_name
        return fitter_records.loc[fitter_records['name'] == name, 'Mean_rank'].values[0]


class NetworkFitter:

    """
    Class used to fit a Fracture or Fracture network object

    :param obj: fracture/fracture network object
    :param use_survival: Boolean flag to use survival (True) or treat the data as if there were no censoring (False). Default is True.
    :param complete_only: Boolean flag to use only complete measurements (True) or all the dataset (False). This flag is used only when use_survival is False. Default is False.
    :param use_AIC: Boolean flag to use AIC (True) or AICc (False) for model selection. Default is True. The
    column name in the dataframe will remain the same (Akaike).

    Add bool to:
        + Consider censored lengths as non-censored
        + Do not consider censored lengths at all

    """
    def __init__(self, obj=None, use_survival=True, complete_only=False, use_AIC=True):

        self._net_data: NetworkData
        self._accepted_fit: list = []
        self._rejected_fit: list = []
        self._AIC_flag: bool = use_AIC
        self._fit_dataframe: DataFrame = DataFrame(columns=['name', 'Akaike',
                                                            'delta_i', 'w_i',
                                                            'max_log_likelihood', 'KS_distance',
                                                            'KG_distance', 'AD_distance',
                                                            'Akaike_rank', 'KS_rank',
                                                            'KG_rank', 'AD_rank',
                                                            'Mean_rank', 'distribution'])

        self.network_data = NetworkData(obj, use_survival, complete_only)

    @property
    def network_data(self) -> NetworkData:
        """
        Property that returns or sets a NetworkData object
        :return:
        """
        return self._net_data

    @network_data.setter
    def network_data(self, data: NetworkData):
        self._net_data = data

    def fit(self, distribution_name: str):

        """
        Fit the data of the entity_df using scipy available distributions
        :param distribution_name: Name of the distribution to fit
        :return:
        """
        print(f'Fitting {distribution_name} on data')
        last_pos = len(self._fit_dataframe)  # The position of a new entry in the dataframe will be the last (i.e. the length of the dataframe)

        self._fit_dataframe.loc[last_pos, 'name'] = distribution_name

        scipy_distribution = getattr(ss, distribution_name)

        if distribution_name == 'norm' or distribution_name == 'logistic':
            params = scipy_distribution.fit(self.network_data.data)
        else:
            params = scipy_distribution.fit(self.network_data.data, floc=0)

        distribution = NetworkDistribution(parent=self, obj=scipy_distribution,
                                           parameters=params, fit_data=self.network_data)

        if self._AIC_flag:
            akaike = distribution.AIC
        else:
            akaike = distribution.AICc

        self._fit_dataframe.loc[last_pos, 'Akaike'] = akaike

        log_likelihood = distribution.max_log_likelihood

        self._fit_dataframe.loc[last_pos, 'max_log_likelihood'] = log_likelihood
        for i, AIC_val in enumerate(self._fit_dataframe['Akaike']):
            d_i = AIC_val - min(self._fit_dataframe['Akaike'])
            self._fit_dataframe.loc[i, 'delta_i'] = d_i
        delta_values = (-self._fit_dataframe['delta_i'].values/2).astype(float)
        total = exp(delta_values).sum()

        for d, delta_i in enumerate(self._fit_dataframe['delta_i']):
            w_i = np.round(exp(-delta_i/2)/total, 5)
            self._fit_dataframe.loc[d, 'w_i'] = w_i

        self._fit_dataframe.loc[last_pos, 'KS_distance'] = distribution.KS_distance
        self._fit_dataframe.loc[last_pos, 'KG_distance'] = distribution.KG_distance
        self._fit_dataframe.loc[last_pos, 'AD_distance'] = distribution.AD_distance

        self._fit_dataframe['Akaike_rank'] = ss.rankdata(self._fit_dataframe['Akaike']).astype(int)
        self._fit_dataframe['KS_rank'] = ss.rankdata(self._fit_dataframe['KS_distance']).astype(int)
        self._fit_dataframe['KG_rank'] = ss.rankdata(self._fit_dataframe['KG_distance']).astype(int)
        self._fit_dataframe['AD_rank'] = ss.rankdata(self._fit_dataframe['AD_distance']).astype(int)
        self._fit_dataframe['Mean_rank'] = self._fit_dataframe.iloc[:, 8:12].mean(axis=1)

        self._fit_dataframe.loc[last_pos, 'distribution'] = distribution

        # self._fit_dataframe.loc[last_pos, 'params'] = params  # this gives out an error for setting the df, I do not know why

    def fit_records(self, sort_by='Akaike') -> DataFrame:

        """ Return the sorted fit dataframe"""

        return self._fit_dataframe.sort_values(by=sort_by, ignore_index=True)

    def get_fitted_distribution(self, distribution_name: str, sort_by='Akaike') -> NetworkDistribution:

        """
        get the fitted NetworkDistribution object
        :param distribution_name: name of the distribution
        :param sort_by: Column name to sort the output order
        :return:
        """
        fit_records = self.fit_records(sort_by)
        distribution = fit_records.loc[fit_records['name'] == distribution_name, 'distribution'].values[0]

        return distribution

    def get_fitted_distribution_names(self, sort_by='Akaike') -> list:

        """
        get the names of the fitted NetworkDistribution object
        :param sort_by: Column name to sort the output order
        :return:
        """

        return self.fit_records(sort_by)['name'].values

    def get_fitted_distribution_list(self, distribution_names: list = None, sort_by='Akaike') -> list:
        """
        Get a list of NetworkDistribution objects for the given distribution name
        :param distribution_names:
        :param sort_by: Column name to sort the output order
        :return:
        """
        fit_records = self.fit_records(sort_by)
        if distribution_names is None:
            distribution_names = fit_records['name'].tolist()

        distribution_list = []

        for name in distribution_names:
            distribution = fit_records.loc[fit_records['name'] == name, 'distribution'].values[0]
            distribution_list.append(distribution)

        return distribution_list

    def get_fitted_parameters(self, distribution_name: str, sort_by='Akaike') -> tuple:
        """
        Get the fitted distributions parameters in the fit records df
        :param distribution_name: Name of the distribution
        :param sort_by: Column name to sort the output order
        """
        fit_records = self.fit_records(sort_by)
        dist = fit_records.loc[fit_records['name'] == distribution_name, 'distribution'].values[0]
        parameters = dist.distribution_parameters
        return parameters

    def get_fitted_parameters_list(self, distribution_names: list = None, sort_by='Akaike') -> list:

        """
        Get the parameters of the computed fit(s)
        :return: Pandas DataFrame
        """

        fit_records = self.fit_records(sort_by)

        if distribution_names is None:
            distribution_names = fit_records['name'].tolist()

        parameter_list = []

        for name in distribution_names:
            parameter_list.append(self.get_fitted_parameters(name))

        return parameter_list

    def best_fit(self, sort_by='Akaike') -> pd.Series:

        """
        Return the best fit in the fit records dataframe sorted by sort_by
        :return:
        """

        df = self.fit_records(sort_by)

        return df.loc[0]

    # ====================== Plot ==========================

    def plot_PIT(self,  show_plot: bool = True,
                 position: list = None, sort_by: str = 'Akaike',
                 bw: bool = False,
                 second_axis: bool = True,
                 n_ticks: int = 5):
        """
        Method to plot the uniform comparison plot for the fitted models.

        :param show_plot: Bool. If True, show the plot if False do not show. Default is True.
        :param position: List. Plot the models at the given position in the ordered fit_records dataframe. If None (default) plot all models.
        :param sort_by: String. Column name to sort the fit_records dataframe. Default is Akaike
        :param bw: Bool. If true turn the plot color-blind friendly (black lines with different patterns). Max 7 lines (i.e. models). Default is False
        :param second_axis: Add the secondary x-axis to the PIT plot
        :param n_ticks: Number of ticks for the secondary x-axis
        :return:
        """

        plotter.matplot_stats_uniform(self, show_plot, position, sort_by, bw, second_axis, n_ticks)

    def tick_plot(self, show_plot: bool = True,
                  position: list = None, n_ticks: int = 5,
                  sort_by: str = 'Akaike'):
        """
        Method used to plot the tick plot of the given fitted model(s)
        :param show_plot: Bool. If True, show the plot if False do not show. Default is True.
        :param position: List. Plot the models at the given position in the ordered fit_records dataframe. If None (default) plot all models.
        :param n_ticks: Number of ticks per bar
        :param sort_by: String. Column name to sort the fit_records dataframe. Default is Akaike
        :return:
        """

        plotter.matplot_tick_plot(self, show_plot=show_plot, position=position, n_ticks=n_ticks, sort_by=sort_by)

    def plot_summary(self, show_plot:bool = True, position: list = None, sort_by: str = 'Akaike'):
        """
        Method to plot the summary plot for the given fitted model(s).
        :param show_plot: Bool. If True, show the plot if False do not show. Default is True.
        :param position: List. Plot the models at the given position in the ordered fit_records dataframe. If None (default) plot all models.
        :param sort_by: String. Column name to sort the fit_records dataframe. Default is Akaike
        :return:
        """

        plotter.matplot_stats_summary(self, show_plot=show_plot, position=position, sort_by=sort_by)

    # ====================== Export ==========================
    def fit_result_to_csv(self, path, sort_by='Akaike'):
        """
        Save the csv to a specified path
        :param path: Path of where to save the csv
        :param sort_by: Column name to sort the values
        :return:
        """
        self.fit_records(sort_by).iloc[:, :-1].to_csv(path, sep=';', index=False)

    def fit_result_to_excel(self, path, sort_by='Akaike'):
        """
        Save the excel to a specified path
        :param path: Path of where to save the excel
        :param sort_by: Column name to sort the values
        :return:
        """
        self.fit_records(sort_by).iloc[:, :-1].to_excel(path, index=False)

    def fit_result_to_clipboard(self, sort_by='Akaike'):
        """
        Copy to clipboard the df to be easily pasted in an excel file
        :param sort_by: Column name to sort the values
        :return:
        """
        self.fit_records(sort_by).iloc[:, :-1].to_clipboard(index=False)

    # def plot_kde(self, n_bins: int = 25, x_min: float = 0.0, x_max: float = None, res: int = 1000):
    #
    #     """
    #     Plot the Kernel Density Estimation PDF function togather with the data histogram
    #
    #     Parameters
    #     -------
    #     n_bins: Number of histogram bins
    #
    #     x_min: Lower value of the range. By default is set to 0
    #
    #     x_max: Higher value of the range. If None the maximum length of the dataset is used. None by default
    #
    #     res: Point resolution between x_min and x_max. By default is set to 1000
    #
    #     """
    #
    #     data = self._lengths
    #
    #     if x_max is None:
    #         x_max = self._lengths.max()
    #
    #     x_vals = np.linspace(x_min, x_max, res)
    #
    #     pdf_kde = gaussian_kde(data).evaluate(x_vals)
    #
    #     fig = plt.figure(num='Gaussian KDE estimation')
    #     plt.hist(data, density=True, bins=15,label='Data')
    #     plt.plot(x_vals, pdf_kde, label='Gaussian KDE')
    #
    #     plt.title('Gaussian KDE estimation')
    #     plt.xlabel('Length [m]')
    #     plt.ylabel('Function response')
    #     plt.grid(True)
    #     plt.show()


"""
In general, it would be good to give the possibility to treat separately sets of fractures/different types of nodes/
boundary and then everything together to consider the network as a composition of single entities and not as a single
entity.

todo make somehow nodes and fractures connected so that for example when only a set is displayed only the nodes of the corresponding set are considered
"""
import os.path

import numpy as np
from geopandas import GeoDataFrame, GeoSeries, read_file
import pandas as pd
from pandas import DataFrame
from shapely.geometry import MultiLineString, Polygon, LineString, Point, MultiPoint
from pyvista import PolyData, DataSet, wrap
from networkx import Graph
from vtkmodules.vtkFiltersCore import vtkConnectivityFilter

import fracability.Plotters as plts

import fracability.Adapters as Rep
from fracability.AbstractClasses import BaseEntity
from fracability.operations import Geometry, Topology


class Nodes(BaseEntity):
    """
    Node base entity, represents all the nodes in the network.

    Parameters
    -----------
    gdf: GeoDataFrame
        Use as input a geopandas dataframe
    csv: str
        Use as input a csv indicated by the path
    shp: str
        Use as input a shapefile indicated by the path
    node_type: int
        Node type. I:1, Y:3, X:4, U:5

    Notes
    --------
    The csv needs to have a "geometry" column. If missing the import will fail.
    """

    def __init__(self, gdf: GeoDataFrame = None, csv: str = None, shp: str = None, node_type: int = -9999):

        self.node_type = node_type
        if gdf is not None:
            super().__init__(gdf=gdf)
        elif csv is not None:
            super().__init__(csv=csv)
        elif shp is not None:
            super().__init__(shp=shp)

    @property
    def entity_df(self) -> GeoDataFrame:
        return self._df

    @entity_df.setter
    def entity_df(self, gdf: GeoDataFrame):
        """
        Each entity process the input dataframe in different ways.
        Nodes modify the entity_df only to add the type column (if not
        already present)
        """

        self._df = gdf
        columns = self._df.columns
        if 'og_line_id' not in columns:
            self._df['og_line_id'] = np.array(gdf.index.values+1)
        if 'type' not in columns:
            self._df['type'] = 'node'
        if 'n_type' not in columns:
            self._df['n_type'] = self.node_type

    @property
    def vtk_object(self) -> PolyData:

        df = self.entity_df
        vtk_obj = Rep.node_vtk_rep(df)
        return vtk_obj

    @vtk_object.setter
    def vtk_object(self, obj: DataSet):
        for index, point in enumerate(obj.points):
            self.entity_df.loc[self.entity_df['id'] == index, 'geometry'] = Point(point)

    def network_object(self) -> Graph:
        network_obj = Rep.networkx_rep(self.vtk_object)
        return network_obj

    @property
    def ternary_node_count(self) -> tuple[float, float, float, float, float]:
        """
        Calculate the node proportions and precise connectivity value following Manzocchi 2002
        :return: A tuple of values PI, PY, PX, PU, CI
        """
        count_dict = self.node_count

        if 1 in count_dict.keys():
            I_nodes = count_dict[1]
        else:
            I_nodes = 0

        if 3 in count_dict.keys():
            Y_nodes = count_dict[3]
        else:
            Y_nodes = 0

        if 4 in count_dict.keys():
            X_nodes = count_dict[4]
        else:
            X_nodes = 0

        if 5 in count_dict.keys():
            U_nodes = count_dict[5]
        else:
            U_nodes = 0

        if 6 in count_dict.keys():
            Y2_nodes = count_dict[6]
        else:
            Y2_nodes = 0

        tot_Y_nodes = Y_nodes + Y2_nodes

        if I_nodes + tot_Y_nodes == 0:
            PI = 0
        else:
            PI = 100 * I_nodes / (I_nodes + tot_Y_nodes + X_nodes)

        if tot_Y_nodes + X_nodes == 0:
            PY = 0
        else:
            PY = 100 * tot_Y_nodes / (I_nodes + tot_Y_nodes + X_nodes)

        if I_nodes + X_nodes == 0:
            PX = 0
        else:
            PX = 100 * X_nodes / (I_nodes + tot_Y_nodes + X_nodes)

        if I_nodes + U_nodes == 0:
            PU = 0
        else:
            PU = 100 * U_nodes / (I_nodes + tot_Y_nodes + X_nodes)

        precise_n = 4 * (1 - PI / 100) / (1 - PX / 100)

        return PI, PY, PX, PU, precise_n

    @property
    def node_count(self) -> dict:
        """
        Calculate the node count dictionary

        :return: A dictionary of nodes and their count
        """

        nodes = self.vtk_object['n_type']
        unique, count = np.unique(nodes, return_counts=True)
        count_dict = dict(zip(unique, count))
        return count_dict
    @property
    def n_censored(self) -> int:
        """ Return the number of censored nodes"""

        nodes = self.vtk_object['n_type']
        unique, count = np.unique(nodes, return_counts=True)
        count_dict = dict(zip(unique, count))

        if 5 not in count_dict.keys():
            return 0
        else:
            return count_dict[5]

    @property
    def n_complete(self) -> int:
        """ Return the number of I nodes"""

        nodes = self.vtk_object['n_type']
        unique, count = np.unique(nodes, return_counts=True)
        count_dict = dict(zip(unique, count))

        return count_dict[1]

    def node_origin(self, node_type: int) -> GeoSeries:
        """
        Return the node origin for the given node type (i.e. which set/sets is/are associated to the node)
        :param node_type:
        :return:
        """
        return self.entity_df.loc[self.entity_df['n_type'] == node_type, 'n_origin']

    def mat_plot(self, markersize=7, return_plot=False, show_plot=True):
        """
        Plot Nodes object with matplot
        :param markersize:
        :param return_plot:
        :param show_plot:
        :return:
        """
        plts.matplot_nodes(self, markersize, return_plot, show_plot)

    def vtk_plot(self, markersize=7, return_plot=False, show_plot=True, notebook=True):
        """
        Plot Nodes object with VTK
        :param markersize:
        :param return_plot:
        :param show_plot:
        :return:
        """

        plts.vtkplot_nodes(self, markersize, return_plot, show_plot, notebook=notebook)

    def ternary_plot(self):
        plts.matplot_ternary(self)


class Fractures(BaseEntity):
    """
    Base entity for fractures, represents all the fractures in the network.

    Parameters
    -----------
    gdf: GeoDataFrame
        Use as input a geopandas dataframe
    csv: str
        Use as input a csv indicated by the path
    shp: str
        Use as input a shapefile indicated by the path
    set_n: int
        Fracture set number.
    check_geometry: Bool
        Perform geometry check. Default is False


    Notes
    --------
    + The csv needs to have a "geometry" column. If missing the import will fail.
    + If the input csv or shapefile has an f_set column then the set_n will be ignored
    """

    def __init__(self, gdf: GeoDataFrame = None, csv: str = None,
                 shp: str = None, set_n: int = None,
                 check_geometry: bool = False):

        self.check_geometries_flag = check_geometry
        self._set_n = set_n
        if gdf is not None:
            super().__init__(gdf=gdf)
        elif csv is not None:
            super().__init__(csv=csv)
        elif shp is not None:
            super().__init__(shp=shp)
        else:
            super().__init__()

    @property
    def set_n(self) -> int:
        """
        Get the set number
        """
        return self._set_n

    @property
    def entity_df(self):
        return self._df

    @entity_df.setter
    def entity_df(self, gdf: GeoDataFrame):
        """
        Each entity process the input dataframe in different ways.
        Fractures modify the entity_df in a couple of ways:
            + No Multilines can be used.
            + If not f_set column is present, it will be created following the set_n value
            + If no length column is present, it will be created (with length rounded to the 4th decimal point)
            + If no censoring column is present, it will be created setting all values to 0
        """
        multiline_list = []
        for index, geom in zip(gdf.index, gdf['geometry']):  # For each geometry in the df

            if isinstance(geom, MultiLineString):
                multiline_list.append(index)
                continue
        if len(multiline_list) > 0:
            print(f'Multilines found, removing from database. If necessary correct them: {np.array(multiline_list)+1}')

        gdf.drop(multiline_list, inplace=True)

        self._df = gdf.copy()
        self._df.reset_index(inplace=True, drop=True)
        columns = self._df.columns
        if 'og_line_id' not in columns:
            self._df['og_line_id'] = np.array(gdf.index.values+1)
        if 'type' not in columns:
            self._df['type'] = 'fracture'
        if 'censored' not in columns:
            self._df['censored'] = 0
        if 'f_set' not in columns:  # todo change this to "set_n" and check at least for similar names
            self._df['f_set'] = self.set_n
        if 'length' not in columns:
            self._df['length'] = np.round(self._df['geometry'].length, 4)

        if self.check_geometries_flag:
            self.remove_double_points()
            self.check_geometries()

    @property
    def vtk_object(self) -> PolyData:
        df = self.entity_df
        vtk_obj = Rep.frac_vtk_rep(df)
        return vtk_obj

    @vtk_object.setter
    def vtk_object(self, obj: DataSet):
        if 'RegionId' not in obj.array_names:
            regions_ids = np.arange(0, obj.n_cells)
            obj['RegionId'] = regions_ids

        if not self.entity_df.empty:
            for region_id in set(obj['RegionId']):
                region = obj.extract_points(obj['RegionId'] == region_id)
                self.entity_df.loc[self.entity_df['id'] == region, 'geometry'] = LineString(region.points)
        else:
            idx = list(set(obj['RegionId']))
            geometry = []
            for region_id in idx:
                region = wrap(obj.extract_cells(obj['RegionId'] == region_id))
                index_list, ind = np.unique(region.cell_connectivity, return_index=True)
                index_list = index_list[np.argsort(ind)]
                geometry.append(LineString(region.points[index_list]))
            d = {'id': idx, 'geometry': geometry}
            gdf = GeoDataFrame(d)
            self.entity_df = gdf

    def network_object(self) -> Graph:
        network_obj = Rep.networkx_rep(self.vtk_object)
        return network_obj

    def check_geometries(self, remove_dup=True, save_shp=False):
        """
        Method used to check if the geometries are correct i.e.:
        + No repeating points
        + No overlaps

        By default, the method will return a list of geometries that need to be fixed. Additionally, a shp file can be
        saved with only the geometries that need to be corrected.

        :param remove_dup: Automatically remove duplicate points. By default, True
        :param save_shp: Save in the same folder of the input shp with only the geometries that need to be corrected.
                         This is a useful support file to be imported in gis to quickly find the problematic geometries.
                         False by default. todo to be implemented
        """

        overlaps_list = []
        tot_geom = len(self.entity_df.geometry)
        print('\n\n')
        for line, geom in enumerate(self.entity_df.geometry):
            print(f'Checking geometries: {line}/{tot_geom}',end='\r')
            if geom is None:
                print(f"\n\nWarning, empty geometry at line {line+1}, fix in GIS\n\n")
            overlaps = self.entity_df.overlaps(geom)
            if overlaps.any():
                overlaps_list.append(self.entity_df.loc[line, 'og_line_id'])
        if len(overlaps_list) > 0:
            print(f'\n\nDetected overlaps for set {self._set_n}: {overlaps_list}. Check geometries in gis and fix.\n\n')

    def mat_plot(self,
                 linewidth=1,
                 color='black',
                 color_set=False,
                 return_plot=False,
                 show_plot=True):
        """
        Plot fracture object with matplotlib

        :param linewidth:
        :param color:
        :param color_set:
        :param return_plot:
        :param show_plot:
        :return:
        """

        plts.matplot_fractures(self,
                               linewidth,
                               color,
                               color_set,
                               return_plot,
                               show_plot)

    def vtk_plot(self,
                 linewidth=1,
                 color='black',
                 color_set=False,
                 return_plot=False,
                 show_plot=True,
                 display_property: str = None,
                 notebook=True):
        """
        Plot fracture object with VTK
        :param linewidth:
        :param color:
        :param color_set:
        :param return_plot:
        :param show_plot:
        :param display_property:
        :return:
        """
        plts.vtkplot_fractures(self,
                               linewidth=linewidth,
                               color=color,
                               color_set=color_set,
                               return_plot=return_plot,
                               show_plot=show_plot,
                               display_property=display_property,
                               notebook=notebook)


class Boundary(BaseEntity):

    """
    Base entity for boundaries, represents all the boundaries in the network.

    Parameters
    -----------
    gdf: GeoDataFrame
        Use as input a geopandas dataframe
    csv: str
        Use as input a csv indicated by the path
    shp: str
        Use as input a shapefile indicated by the path
    group_n: int
        Boundary number.
    check_geometry: Bool
        Perform geometry check. Default is False



    Notes
    --------
    + The csv needs to have a "geometry" column. If missing the import will fail.
    """
    def __init__(self, gdf: GeoDataFrame = None,
                 csv: str = None, shp: str = None,
                 group_n: int = 1, check_geometry: bool = False):
        """
        Init for Boundary entity. Different inputs can be used. Geopandas dataframe, csv or shapefile. The csv needs to be
        structured in such a way to be compatible with the Nodes entity.

        :param gdf: Geopandas dataframe
        :param csv: Path of a csv
        :param shp: Path of the shapefile
        :param group_n: Boundary group number
        """

        self.group_n = group_n
        self.check_geometries_flag = check_geometry
        if gdf is not None:
            super().__init__(gdf=gdf)
        elif csv is not None:
            super().__init__(csv=csv)
        elif shp is not None:
            super().__init__(shp=shp)


    @property
    def entity_df(self):
        return self._df

    @entity_df.setter
    def entity_df(self, gdf: GeoDataFrame):
        """
        Each entity process the input dataframe in different ways.
        Boundaries modify the entity_df by converting Polygons in Linestrings
        to (using the boundary method) and MultiLinestrings to LineStrings.
        A 'type' column is added if missing.
        """

        self._df = gdf
        self._df.reset_index(inplace=True, drop=True)

        geom_list = []
        # boundaries = df.boundary
        #
        # gdf = boundaries.explode(ignore_index=True)

        # The following is horrible and I hate it but for some reason the commented lines above
        # do not work for shapely 1.8 and geopandas 0.11 while they work perfectly with 2.0 and 0.13
        # When PZero moves to shapely 2.0 remove the lines between these comments
        # and uncomment the two lines above

        # Remove from here
        # This is to suppress the SettingWithCopyWarning (we are not working on a copy)
        pd.options.mode.chained_assignment = None

        for index, line in enumerate(self._df.loc[:, 'geometry']):
            if isinstance(line, Polygon):
                bound = line.boundary
                if isinstance(bound, MultiLineString):
                    for linestring in bound.geoms:
                        geom_list.append(linestring)
                else:
                    geom_list.append(bound)
            else:
                geom_list.append(line)

        for index, value in enumerate(geom_list):
            self._df.loc[index, 'geometry'] = value
        # Remove up to here

        columns = self._df.columns
        if 'og_line_id' not in columns:
            self._df['og_line_id'] = np.array(gdf.index.values+1)
        if 'type' not in columns:
            self._df['type'] = 'boundary'
        if 'b_group' not in columns:
            self._df['b_group'] = self.group_n

        if self.check_geometries_flag:
            self.remove_double_points()

    @property
    def vtk_object(self) -> PolyData:

        df = self.entity_df
        vtk_obj = Rep.bound_vtk_rep(df)
        return vtk_obj

    @vtk_object.setter
    def vtk_object(self, obj: DataSet):

        if 'RegionId' not in obj.array_names:
            regions_ids = np.arange(0, obj.n_cells)
            obj['RegionId'] = regions_ids

        if not self.entity_df.empty:
            for region_id in set(obj['RegionId']):
                region = obj.extract_points(obj['RegionId'] == region_id)
                points = region.points
                if points[0] != points[-1]:  # All boundaries must be closed
                    points = np.append(points, [points[0]], axis=0)

                self.entity_df.loc[self.entity_df['id'] == region, 'geometry'] = LineString(points)
        else:
            idx = list(set(obj['RegionId']))
            geometry = []
            for region_id in idx:
                region = obj.extract_cells(obj['RegionId'] == region_id)
                points = region.points
                if np.any(points[0, :] != points[-1, :]):  # All boundaries must be closed
                    points = np.append(points, [points[0]], axis=0)
                geometry.append(LineString(points))

            d = {'id': idx, 'geometry': geometry}
            gdf = GeoDataFrame(d)
            self.entity_df = gdf

    def network_object(self) -> Graph:
        network_obj = Rep.networkx_rep(self.vtk_object)
        return network_obj

    def mat_plot(self,
                 linewidth=1,
                 color='red',
                 return_plot=False,
                 show_plot=True):
        """
        Plot Boundary object with matplotlib
        :param linewidth:
        :param color:
        :param return_plot:
        :param show_plot:
        :return:
        """

        plts.matplot_boundaries(self,
                                linewidth,
                                color,
                                return_plot,
                                show_plot)

    def vtk_plot(self,
                 linewidth=1,
                 color='red',
                 color_set=False,
                 return_plot=False,
                 show_plot=True,
                 notebook=True):
        """
        Plot Boundary object with vtk
        :param linewidth:
        :param color:
        :param color_set:
        :param return_plot:
        :param show_plot:
        :return:
        """
        plts.vtkplot_boundaries(self,
                                linewidth=linewidth,
                                color=color,
                                return_plot=return_plot,
                                show_plot=show_plot,
                                notebook=notebook)


class Backbone(Fractures):
    """
    Base entity for the backbone. It inherits the same properties of Fractures and is made to differentiate the two types and in case add backbone specific methods/properties if needed
    """
    def __init__(self, gdf: GeoDataFrame = None, csv: str = None,
                 shp: str = None, set_n: int = None,
                 check_geometry: bool = False):
        super().__init__(gdf, csv, shp, set_n, check_geometry)


class FractureNetwork(BaseEntity):
    """
    Fracture network base entity. Fracture networks are defined by one or
    more:

        + Fracture base entities

        + Boundary base entities

        + Nodes base entities

    All the data is represented in the entity_df and the different objects
    are defined by the 'type' column.

    FractureNetwork objects can be created in two ways depending on how
    the dataset is structured.

        1. If fractures and boundaries and nodes are saved in different shp files
        then use the add_fracture,add_boundary and add_nodes methods on an empty
        FractureNetwork object.

        2. If fractures and boundaries and nodes are saved in a single shp the
        geopandas dataframe can be directly fed when instantiating the class.
        In this case a type column must be set to indicate of which type the geometries are

    """

    def __init__(self, gdf: GeoDataFrame = None, csv: str = None):
        """
        Init for fracture network entity. Different inputs can be used. Geopandas dataframe, csv or shapefile.
        The csv needs to be structured in such a way to be compatible with the Nodes entity.

        :param gdf: Geopandas dataframe
        :param csv: Path of a csv
        """

        self.column_names = ['type', 'object', 'n_type', 'f_set', 'b_group', 'active']
        self._df: DataFrame = DataFrame(columns=self.column_names)

        if csv is not None:
            gdf = read_file(csv, GEOM_POSSIBLE_NAMES="geometry", KEEP_GEOM_COLUMNS="NO")

        if gdf is not None:
            nodes = Nodes(gdf.loc[gdf['type'] == 'node'])
            fractures = Fractures(gdf.loc[gdf['type'] == 'fracture'])
            boundaries = Boundary(gdf.loc[gdf['type'] == 'boundary'])

            self.add_nodes(nodes)
            self.add_fractures(fractures)
            self.add_boundaries(boundaries)

    @property
    def entity_df(self):
        return self._df

    @entity_df.setter
    def entity_df(self, gpd: GeoDataFrame):

        nodes = gpd.loc[gpd['type'] == 'nodes']

        if not nodes.empty:

            types = set(nodes['node_type'])

            for node_type in types:
                nodes = Nodes(gdf=nodes.loc[nodes['n_type'] == node_type], node_type=node_type)
                if self.is_type_active(node_type):
                    self.add_nodes(nodes)

        fractures = gpd.loc[gpd['type'] == 'fracture']

        if not fractures.empty:
            sets = set(fractures['f_set'])

            for set_n in sets:
                fracture = Fractures(gdf=fractures.loc[fractures['f_set'] == set_n], set_n=set_n)
                if self.is_set_active(set_n):
                    self.add_fractures(fracture)

        boundaries = gpd.loc[gpd['type'] == 'boundary']

        if not boundaries.empty:

            groups = set(boundaries['b_group'])

            for group_n in groups:
                boundary = Boundary(gdf=boundaries.loc[boundaries['b_group'] == group_n], group_n=group_n)
                if self.is_group_active(group_n):
                    self.add_boundaries(boundary)

    @property
    def crs(self):
        return self.fracture_network_to_components_df().crs

    #  ==================== Nodes property ====================

    @property
    def nodes(self) -> Nodes:
        """
        Property that returns a Node entity object of all the active nodes.
        :return: Nodes entity object
        """
        if self._active_nodes_df is not None:
            return Nodes(self._active_nodes_df)
        else:
            return None

    @property
    def _nodes_components(self) -> DataFrame:
        """
        Property that returns the nodes components of the Fracture network df.
        :return: Pandas dataframe slice of the nodes components
        """
        nodes = self.entity_df.loc[self.entity_df['type'] == 'nodes']
        if not nodes.empty:
            return nodes
        else:
            return None

    @property
    def _active_nodes_components(self) -> DataFrame:
        """
        Internal property that returns the active nodes components in the of the Fracture network df
        :return: Pandas dataframe slice of active nodes components
        """
        nodes = self.entity_df.loc[self.entity_df['type'] == 'nodes']

        active_nodes = nodes.loc[nodes['active'] == 1]

        if not active_nodes.empty:
            return active_nodes
        else:
            return None

    @property
    def _active_nodes_df(self) -> GeoDataFrame:
        """
        Internal property that returns the dataframe of the active nodes
        :return:
        """

        gdf = GeoDataFrame()

        nodes = self._active_nodes_components

        if nodes is not None:
            for row, fracture in nodes.iterrows():
                gdf = pd.concat([gdf, fracture['object'].entity_df], ignore_index=True)
            return gdf
        else:
            return None

    def add_nodes(self, nodes: Nodes = None):
        """
        Method used to add nodes components to the fracture network Dataframe
        :param nodes: Nodes object to be added

        Notes
        -------
        The nodes are added to the fracture network dataframe using the assigned node_type present in the
        Nodes object dataframe. If the node type is already present the node object will be overwritten if not it will
        be appended
        """

        node_types = set(nodes.entity_df['n_type'])

        for node_type in node_types:

            nodes_df = nodes.entity_df.loc[nodes.entity_df['n_type'] == node_type]

            nodes_group = Nodes(gdf=nodes_df, node_type=node_type)

            if node_type not in self._df['n_type'].values:
                new_df = DataFrame([['nodes', nodes_group, node_type, 1]], columns=['type', 'object',
                                                                                    'n_type', 'active'])
                self._df = pd.concat([self._df, new_df], ignore_index=True)
            else:
                self._df.loc[self._df['n_type'] == node_type, 'object'] = nodes_group

    def add_nodes_from_dict(self, node_dict, classes=None, origin_dict: dict = None):
        """Add nodes a dict of shapely geometry (key), classes and optionally node origin (value).

        :param node_dict: Dict of shapely node geometries as keys and a list (class, node_index).
        :param classes: List of node classes that are needed to be added. If none are provided all the classes are used [1, 3, 4, 5, 6]
        :param origin_dict: Dict of shapely node geometries as keys and a list of node origin.
        """

        node_geometry = np.array(list(node_dict.keys()))
        class_list = np.array(list(node_dict.values()))
        origin_list = list(origin_dict.values())

        entity_df = GeoDataFrame({'type': 'node', 'n_type': class_list[:, 0], 'n_index': class_list[:, 1],
                                  'n_origin': origin_list, 'geometry': node_geometry}, crs=self.crs)

        nodes = Nodes(gdf=entity_df)

        self.add_nodes(nodes)

    def nodes_object(self, node_type: int) -> Nodes:
        """
        Method that returns the Node object of a given node_type
        :param node_type: Type of the node
        :return: Nodes object
        """
        return self._fractures_components.loc[self._fractures_components['n_type'] == node_type, 'object'].values[0]

    def activate_nodes(self, node_type: list = None):
        """
        Method that activates the nodes provided in the node_type list.
        :param node_type: List of node types to be activated
        """

        if node_type is None:
            self.entity_df.loc[self.entity_df['type'] == 'nodes', 'active'] = 1
        else:
            self.entity_df.loc[self.entity_df['type'] == 'nodes', 'active'] = 0
            for t in node_type:
                self.entity_df.loc[self.entity_df['n_type'] == t, 'active'] = 1

    def is_type_active(self, node_type: int) -> bool:
        """
        Method used to return if a given node type is active in the fracture network
        :param node_type: node type to check
        :return: Bool value of the test
        """
        nodes = self._nodes_components

        value = nodes.loc[nodes['n_type'] == node_type, 'active'].values[0]

        return value

    #  ==================== Fractures  ====================

    @property
    def fractures(self) -> Fractures:
        """
        Property that returns a Fracture entity object of all the active fracture sets.
        :return: Fracture entity object
        """
        if self._active_fractures_df is not None:
            return Fractures(self._active_fractures_df)
        return None

    @property
    def sets(self) -> list:
        """Return the list of the number of sets"""

        sets = list(set(self.fractures.entity_df['f_set'].values))
        return sets

    @property
    def _fractures_components(self) -> DataFrame:
        """
        Internal property that returns the fracture components of the Fracture network df.
        :return: Pandas dataframe slice of the fracture set components
        """
        fractures = self.entity_df.loc[self.entity_df['type'] == 'fractures']

        if not fractures.empty:
            return fractures
        else:
            return None

    @property
    def _active_fractures_components(self) -> DataFrame:
        """
        Internal property that returns the active fracture components in the of the Fracture network df
        :return: Pandas dataframe slice of active fracture components
        """

        fractures = self.entity_df.loc[self.entity_df['type'] == 'fractures']

        active_fractures = fractures.loc[fractures['active'] == 1]
        if not active_fractures.empty:
            return active_fractures
        else:
            return None

    @property
    def _active_fractures_df(self) -> GeoDataFrame:
        """
        Internal property that returns the dataframe of the active fractures sets components
        :return: Geopandas dataframe of the active fracture sets components
        """

        gdf = GeoDataFrame()

        fractures = self._active_fractures_components

        if fractures is not None:
            for row, fracture in fractures.iterrows():
                gdf = pd.concat([gdf, fracture['object'].entity_df], ignore_index=True)

            return gdf
        else:
            return None

    def add_fractures(self, fractures: Fractures = None):
        """
        Method used to add fracture components to the fracture network Dataframe
        :param fractures: Fracture object to be added

        Notes
        -------
        The fractures are added to the fracture network dataframe using the assigned fracture_set present in the
        Fracture object dataframe. If the set is already present the fracture object will be overwritten if not it will
        be appended
        """
        fracture_sets = set(fractures.entity_df['f_set'])

        for set_n in fracture_sets:

            fractures_df = fractures.entity_df.loc[fractures.entity_df['f_set'] == set_n]
            fractures_group = Fractures(gdf=fractures_df, set_n=set_n)

            if set_n not in self._df['f_set'].values:
                new_df = DataFrame([['fractures', fractures_group, set_n, 1]],
                                   columns=['type', 'object', 'f_set', 'active'])
                self._df = pd.concat([self._df, new_df], ignore_index=True)
            else:
                self._df.loc[self._df['f_set'] == set_n, 'object'] = fractures_group

    def fracture_object(self, set_n: int) -> Fractures:
        """
        Method that returns the Fracture object of a given set
        :param set_n: Number of the set
        :return: Fracture object
        """
        return self._fractures_components.loc[self._fractures_components['f_set'] == set_n, 'object'].values[0]

    def activate_fractures(self, set_n: list = None):
        """
        Method that activates the fractures provided in the set_n list.
        :param set_n: List of sets to be activated
        """

        if set_n is None:
            self.entity_df.loc[self.entity_df['type'] == 'fractures', 'active'] = 1
        else:
            self.entity_df.loc[self.entity_df['type'] == 'fractures', 'active'] = 0
            if len(set_n) > 0:
                for n in set_n:
                    self.entity_df.loc[self.entity_df['f_set'] == n, 'active'] = 1

    def is_set_active(self, set_n: int) -> bool:
        """
        Method used to return if a given fracture set is active in the fracture network
        :param set_n: set to check
        :return: Bool value of the test
        """

        fractures = self._fractures_components

        value = fractures.loc[fractures['f_set'] == set_n, 'active'].values[0]

        return value

    #  ==================== Boundaries property ====================

    @property
    def boundaries(self) -> Boundary:
        """
        Property that returns a Boundary entity object of all the active boundary groups.
        :return: Boundary entity object
        """
        if self._active_boundaries_df is not None:
            return Boundary(self._active_boundaries_df)
        else:
            return None

    @property
    def _boundaries_components(self) -> DataFrame:
        """
        Internal property that returns the boundary components of the Fracture network df.
        :return: Pandas dataframe slice of the boundary groups components
        """

        boundaries = self.entity_df.loc[self.entity_df['type'] == 'boundary']
        if not boundaries.empty:
            return boundaries
        else:
            return None

    @property
    def _active_boundaries_components(self) -> DataFrame:
        """
        Internal property that returns the active boundary group components in the of the Fracture network df
        :return: Pandas dataframe slice of active boundary group components
        """

        boundaries = self.entity_df.loc[self.entity_df['type'] == 'boundary']

        active_boundaries = boundaries.loc[boundaries['active'] == 1]

        if not active_boundaries.empty:

            return active_boundaries
        else:
            return None

    @property
    def _active_boundaries_df(self) -> GeoDataFrame:
        """
        Internal property that returns the dataframe of the active boundary
        :return: GeoPandas DataFrame of the active boundaries of the fracture network
        """

        df = GeoDataFrame()

        boundaries = self._active_boundaries_components

        if boundaries is not None:

            for row, fracture in boundaries.iterrows():
                df = pd.concat([df, fracture['object'].entity_df], ignore_index=True)

            return df
        else:
            return None

    def add_boundaries(self, boundary: Boundary = None):
        """
        Method used to add boundary components to the fracture network Dataframe
        :param boundary: Boundary object to be added

        Notes
        -------
        The boundary are added to the fracture network dataframe using the assigned group_n present in the
        Boundary object dataframe. If the group is already present the boundary object will be overwritten if not it will
        be appended
        """

        boundary_groups = set(boundary.entity_df['b_group'])

        for group_n in boundary_groups:

            boundary_df = boundary.entity_df.loc[boundary.entity_df['b_group'] == group_n]
            boundary_group = Boundary(gdf=boundary_df, group_n=group_n)

            if group_n not in self._df['b_group'].values:

                new_df = DataFrame([['boundary', boundary_group, group_n, 1]],
                                   columns=['type', 'object', 'b_group', 'active'])
                self._df = pd.concat([self._df, new_df], ignore_index=True)
            else:
                self._df.loc[self._df['b_group'] == group_n, 'object'] = boundary_group

    def boundary_object(self, group_n: int) -> Boundary:
        """
        Method that returns the Node object of a given group_number
        :param group_n: Number of the group
        :return: Boundary object
        """

        return self._boundaries_components.loc[self._boundaries_components['b_group'] == group_n, 'object'].values[0]

    def activate_boundaries(self, group_n: list = None):

        """
        Method that activates the boundary provided in the group_n list.
        :param group_n: List of groups to be activated
        """

        if group_n is None:
            self.entity_df.loc[self.entity_df['type'] == 'boundary', 'active'] = 1
        else:
            self.entity_df.loc[self.entity_df['type'] == 'boundary', 'active'] = 0
            for n in group_n:
                self.entity_df.loc[self.entity_df['b_group'] == n, 'active'] = 1

    def is_group_active(self, group_n: int) -> bool:
        """
        Method used to return if a given boundary group is active in the fracture network
        :param group_n: set to check
        :return: Bool value of the test
        """

        boundaries = self._boundaries_components
        value = boundaries.loc[boundaries['b_group'] == group_n, 'active'].values[0]

        return value

    #  ==================== Backbone methods ====================

    def calculate_backbone(self, biggest_region=True):
        """
        Calculate the backbone(s) of the network and add the calculated nodes to the network.

        :param biggest_region: Output only most connected region. Default is True
        """

        fractures = self.fractures

        connectivity = vtkConnectivityFilter()

        connectivity.AddInputData(fractures.vtk_object)
        # if biggest_region: # For now only biggest region
        connectivity.SetExtractionModeToLargestRegion()

        connectivity.Update()

        vtkbackbone = PolyData(connectivity.GetOutput())

        backbone_set_n = self.sets[-1] + 1 # use the last available set and add +1. This could be expanded for multiple backbones

        backbone = Backbone(set_n=backbone_set_n)
        backbone.vtk_object = vtkbackbone
        backbone.crs = self.crs

        new_df = DataFrame([['backbone', backbone, backbone_set_n, 0]],
                           columns=['type', 'object', 'f_set', 'active'])
        self._df = pd.concat([self._df, new_df], ignore_index=True)

    @property
    def backbone(self):
        """
        Property that returns a Fractures entity object of the backbone.
        :return: Fracture entity object
        """
        return self._df.loc[self._df['type'] == 'backbone', 'object'].values

    #  ==================== Generic methods ====================

    def fracture_network_to_components_df(self) -> DataFrame:

        """
        Method used to return the fracture network as a single geopandas dataframe.
        :return: Geopandas DataFrame of the whole fracture network
        """

        gdf = DataFrame()

        nodes = self._active_nodes_df

        gdf = pd.concat([gdf, nodes], ignore_index=True)

        fractures = self._active_fractures_df

        gdf = pd.concat([gdf, fractures], ignore_index=True)

        boundaries = self._active_boundaries_df

        gdf = pd.concat([gdf, boundaries], ignore_index=True)

        if 'n_type' in gdf.columns:
            gdf['n_type'] = gdf['n_type'].fillna(-9999).astype('int64')
        gdf['f_set'] = gdf['f_set'].fillna(-9999).astype('int64')
        gdf['censored'] = gdf['censored'].fillna(-9999).astype('int64')
        if 'b_group' in gdf.columns:
            gdf['b_group'] = gdf['b_group'].fillna(-9999).astype('int64')
        return gdf

    def vtk_object(self, include_nodes: bool = True) -> PolyData:

        """
        Method used to return a vtkPolyData representation of the fracture network
        :param include_nodes: Bool flag used to control if include or not the nodes in the fracture network object
        :return: vtkPolyData of the fracture network
        """

        vtk_obj = Rep.fracture_network_vtk_rep(self.fracture_network_to_components_df(), include_nodes=include_nodes)
        return vtk_obj

    def network_object(self) -> Graph:
        """
        Method used to return a networkx Graph representation of the fracture network
        :return: Graph of the fracture network
        """

        network_object = Rep.networkx_rep(self.vtk_object(include_nodes=False))
        return network_object

    def check_network(self, check_single=True, save_shp=None):
        """
        Method used to check if network-wide the geometries are correct i.e.:
        + No repeating points
        + No overlaps

        By default, the method will return a list of geometries that need to be fixed. Additionally, a shp file can be
        saved with only the geometries that need to be corrected.

        :param remove_dup: Automatically remove duplicate points. By default, True. When false, a point shapefile where double points are present will be saved
        :param check_single: Perform check also for the single components
        :param save_shp: Path to save the shp of the check. If check_single is true then also the results of the single component check will be saved.
        """

        df = self.fracture_network_to_components_df()
        df = df.loc[df['type'] != 'node']

        set_n = np.array(list(set(df['f_set'])))

        overlaps_list = []
        overlaps_list_dict = {s: [] for s in set_n[set_n>0]}
        overlaps_geometry_list = []
        intersections_list = []
        intersections_geometry_list = []

        for line, geom in enumerate(df.geometry):

            if df.loc[line, 'type'] == 'boundary':
                mask = np.ones(df.geometry.size, dtype=bool)
                mask[line] = False
                sub_df = df[mask]
                intersections = sub_df.intersects(geom)
                touching = sub_df.touches(geom)
                xor = intersections != touching
                if xor.any():
                    filtered_df = sub_df[xor]
                    for n in set(filtered_df['f_set']):
                        int_dict = dict()
                        int_dict['boundary_int'] = list(filtered_df.loc[filtered_df['f_set'] == n, 'og_line_id'].values)
                        overlaps_list_dict[n].append(int_dict)
                    intersections_list = filtered_df['og_line_id'].values
                    intersections_geometry_list = filtered_df['geometry'].values

            else:
                overlaps = df.overlaps(geom)
                if overlaps.any():
                    overlaps_list.append(df.loc[line, 'og_line_id'])
                    overlaps_geometry_list.append(df.loc[line, 'geometry'])
                    set_n = df.loc[line, 'f_set']
                    overlaps_list_dict[set_n].append(df.loc[line, 'og_line_id'])

        if save_shp:
            path_frac = os.path.join(save_shp, 'frac_corr.shp')

            f_out_dict = {'og_id': [*overlaps_list, *intersections_list],
                          'geometry': [*overlaps_geometry_list, *intersections_geometry_list]}

            out_df = GeoDataFrame(f_out_dict, crs=self.crs)
            out_df.to_file(path_frac)

        else:
            print(overlaps_list_dict)

    def clean_network(self, buffer = 0.05, inplace=True):
        """Tidy the intersection of the active entities in the fracture network. A buffer is applied to all the
        geometries to ensure intersection in a given radius.

        :param buffer: Applied buffer to the geometries of the entity.
        :param inplace: If true automatically replace the network with the clean one, if false then return the clean
         geopandas dataframe. Default is True
         """

        if inplace:
            Geometry.tidy_intersections(self)
        else:
            return Geometry.tidy_intersections(self, inplace=False)

    def calculate_topology(self, clean_network=True):
        """
        Calculate the topology of the network and add the calculated nodes to the network.

        :param clean_network: If true, before calculating the topology the network is cleaned with the clean_network. Default is True
        """
        if clean_network is True:
            self.clean_network()

        nodes_dict, origin_dict = Topology.nodes_conn(self)
        self.add_nodes_from_dict(nodes_dict,origin_dict=origin_dict, classes=None)






    @property
    def fraction_censored(self) -> float:
        """Get the fraction of censored fractures in the network """

        n_censored = self.fractures.entity_df['censored'] == 1

        total = self.fractures.entity_df['censored'] >= 0

        return len(self.fractures.entity_df[n_censored])/len(self.fractures.entity_df[total])

    #  ==================== Plotting methods ====================

    def vtk_plot(self,
                 markersize=5,
                 fracture_linewidth=1,
                 boundary_linewidth=1,
                 fracture_color='black',
                 boundary_color='red',
                 color_set=False,
                 show_plot=True,
                 return_plot=False,
                 notebook=True):
        """
        Method used to plot the fracture network using vtk
        :param markersize:
        :param fracture_linewidth:
        :param boundary_linewidth:
        :param fracture_color:
        :param boundary_color:
        :param color_set:
        :param show_plot:
        :param return_plot:
        :return:
        """

        plts.vtkplot_frac_net(self,
                              markersize=markersize,
                              fracture_linewidth=fracture_linewidth,
                              boundary_linewidth=boundary_linewidth,
                              fracture_color=fracture_color,
                              boundary_color=boundary_color,
                              color_set=color_set,
                              show_plot=show_plot,
                              return_plot=return_plot,
                              notebook=notebook)

    def backbone_plot(self,
                      method='vtk',
                      fracture_linewidth=1,
                      boundary_linewidth=1,
                      fracture_color='black',
                      boundary_color='red',
                      return_plot=False,
                      show_plot=True,
                      notebook=True):
        """
        Method used to plot the fracture network using vtk
        :return:
        """
        if method == 'vtk':
            plts.vtkplot_backbone(self,
                                  fracture_linewidth=fracture_linewidth,
                                  boundary_linewidth=boundary_linewidth,
                                  fracture_color=fracture_color,
                                  boundary_color=boundary_color,
                                  return_plot=return_plot,
                                  show_plot=show_plot,
                                  notebook=notebook)
        elif method == 'matplot':
            plts.matplot_backbone(self,
                                  fracture_linewidth,
                                  boundary_linewidth,
                                  fracture_color,
                                  boundary_color,
                                  return_plot,
                                  show_plot)

    def mat_plot(self,
                 markersize=5,
                 fracture_linewidth=1,
                 boundary_linewidth=1,
                 fracture_color='black',
                 boundary_color='red',
                 color_set=False,
                 show_plot=True,
                 return_plot=False):
        """
        Method used to plot the fracture network using matplotlib
        :param markersize:
        :param fracture_linewidth:
        :param boundary_linewidth:
        :param fracture_color:
        :param boundary_color:
        :param color_set:
        :param show_plot:
        :param return_plot:
        :return:
        """
        plts.matplot_frac_net(self,
                              markersize,
                              fracture_linewidth,
                              boundary_linewidth,
                              fracture_color,
                              boundary_color,
                              color_set,
                              show_plot,
                              return_plot)

    def ternary_plot(self):
        """
        Method used to plot the ternary diagram of the fracture network
        :return:
        """
        plts.matplot_ternary(self)

    #  ==================== Output methods ====================

    def save_csv(self, path: str, sep: str = ',', index: bool = False):
        """
        Save the fracture network entity df as csv
        :param index:
        :type sep: object
        :param path:
        :return:
        """

        if self.nodes is not None:
            self.nodes.save_csv(path)

        if self.fractures is not None:
            self.fractures.save_csv(path)

        if self.boundaries is not None:
            self.boundaries.save_csv(path)

        if self.backbone is not None:
            for bb in self.backbone:
                bb.save_csv(path)

    def save_shp(self, path: str):
        """
        Save the entity df as shp
        :param path:
        :return:
        """

        if self.nodes is not None:
            self.nodes.save_shp(path)

        if self.fractures is not None:
            self.fractures.save_shp(path)

        if self.boundaries is not None:
            self.boundaries.save_shp(path)

        if self.backbone is not None:
            for bb in self.backbone:
                bb.save_shp(path)


"""
FracAbility has different types of data that needs to be represented in different ways. In the Plotters module
different class adapters are proposed to plot the data. It is possible to plot:


Fracture Network (entity):


1. The geopandas dataframe -> matplotlib:
    + Fractures:
        + Rose plot
        + Backbone(s) -> Different colors if multiple backbones are present
    + Boundary
    + Nodes:
        + Different colors and shapes depending on the node type

2. The VTK entities -> pyvista:
    + Fractures:
        + Backbone(s) -> Different colors if multiple backbones are present
    + Boundary
    + Nodes:
        + Different colors depending on the node type

3. The topology data -> matplotlib:
    + I,Y,X node proportions in a ternary plot

Fracture network (statistics):

1. The statistics -> matplotlib:

    + Single distribution plot (each alone or together):
        + pdf
        + cdf
        + sf
        + summary table

    + Multiple distribution plot:
        + cdf vs ecdf
        + pdf vs histograms
        + P-P and Q-Q plot?


todo the plotters should all inherit from a BasePlotter class
"""
import matplotlib
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from pyvista import Plotter
import pyvista as pv
import ternary
from fracability.Statistics import NetworkDistribution, NetworkFitter
from fracability.utils.general_use import KM, setFigLinesBW, ecdf_find_x
import numpy as np


def matplot_nodes(entity,
                  markersize=7,
                  return_plot=False,
                  show_plot=True):
    """
    Plot a fracability Nodes entity using matplotlib.

    :param entity: Nodes entity to plot
    :param markersize: Size of the markers as int
    :param return_plot: Bool. If true the plot is returned. By default, False
    :param show_plot: Bool. If true the plot is shown. By default, True
    :return: If return_plot is true a matplotlib axis is returned

    Notes
    -------
    The nodes are represented using this color and marker codes:

    1. I nodes: blue red circle
    2. Y nodes: green triangle
    3. Y2 nodes: cyan triangle
    4. X nodes: blue square
    5. U nodes: yellow pentagon
    """

    if 'Fracture net plot' not in plt.get_figlabels():
        figure = plt.figure(num=f'Nodes plot')

        ax = plt.subplot(111)
    else:
        ax = plt.gca()
    vtk_object = entity.vtk_object
    points = vtk_object.points
    node_types = vtk_object['n_type']
    I = np.where(node_types == 1)
    Y = np.where(node_types == 3)
    X = np.where(node_types == 4)
    U = np.where(node_types == 5)
    Y2 = np.where(node_types == 6)

    # todo change the colors to reflect the pallete of the paper
    ax.plot(points[I][:, 0], points[I][:, 1], 'or', markersize=markersize)
    ax.plot(points[Y][:, 0], points[Y][:, 1], '^g', markersize=markersize)
    ax.plot(points[Y2][:, 0], points[Y2][:, 1], '^c', markersize=markersize)
    ax.plot(points[X][:, 0], points[X][:, 1], 'sb', markersize=markersize)
    ax.plot(points[U][:, 0], points[U][:, 1], 'py', markersize=markersize)

    if return_plot:
        return ax
    else:
        if show_plot:
            plt.show()


def matplot_ternary(entity,
                    return_plot: bool = False,
                    show_plot: bool = True):

    """
    Plot the ternary diagram for nodes

    Parameters
    ----------

    entity: The fracability Nodes entity

    return_plot: Bool. If true return the plot. By default, false

    show_plot: Bool. If true show the plot. By default, true.


    """

    figure, tax = ternary.figure(scale=100)
    figure.set_size_inches(10, 10)

    if entity.name == 'FractureNetwork':
        nodes = entity.nodes
    elif entity.name == 'Nodes':
        nodes = entity

    PI, PY, PX, _, CI = nodes.ternary_node_count
    points = [(PX, PI, PY)]

    tax.scatter(points, marker='o', color='red', label='Classification')
    tax.annotate(f'{np.round(CI,2)}', position=[PX, PI, PY])

    for n in range(8):
        n += 1
        A1 = np.array([[1, 1, 1], [0, 0, 1], [-n, 4, 0]])
        B = np.array([1, 0, 4 - n])

        X1 = np.linalg.inv(A1).dot(B) * 100
        if n < 4:
            side = [1, 0, 0]
        else:
            side = [0, 1, 0]
        A2 = np.array([[1, 1, 1], side, [-n, 4, 0]])

        X2 = np.linalg.inv(A2).dot(B) * 100

        tax.line(X1, X2, color='black', linewidth=1, alpha=n / 8)

    ax = tax.get_axes()
    ax.text(76.8, 21.3, 8)
    ax.text(74.8, 23.8, 7)
    ax.text(73.5, 27.9, 6)
    ax.text(71, 32.3, 5)
    ax.text(69.1, 38, 4)
    ax.text(65.8, 45, 3)
    ax.text(62.7, 54, 2)
    ax.text(57, 66.5, 1)

    tax.right_corner_label("X", fontsize=15)
    tax.top_corner_label("I", fontsize=15)
    tax.left_corner_label("Y", fontsize=15)

    tax.get_axes().set_aspect(1)  # This is used to avoid deformation when rescaling the plotter window
    # tax.clear_matplotlib_ticks()
    tax.get_axes().axis('off')
    if return_plot:
        return tax
    else:
        if show_plot:
            plt.show()


def matplot_fractures(entity,
                      linewidth=1,
                      color='black',
                      color_set=False,
                      return_plot=False,
                      show_plot=True):
    """
    Plot a fracability Fracture entity using matplotlib.

    :param entity: Fracture entity to plot
    :param linewidth: Size of the lines as int
    :param color: General color of the lines as str.
    :param color_set: Bool. If true the lines are based on the set values.
    :param return_plot: Bool. If true the plot is returned. By default, False
    :param show_plot: Bool. If true the plot is shown. By default, True
    :return: If return_plot is true a matplotlib axis is returned

    """

    if return_plot:
        ax = plt.gca()
    else:
        figure = plt.figure(num=f'Fractures plot')

        ax = plt.subplot(111)

    if color_set:
        if 'f_set' in entity.entity_df.columns:
            n_sets = len(set(entity.entity_df['f_set']))
            cmap = matplotlib.colormaps.get_cmap("rainbow").resampled(n_sets)
            entity.entity_df.plot(ax=ax,
                                  cmap=cmap,
                                  linewidth=linewidth)
        else:
            return False
    else:

        entity.entity_df.plot(ax=ax, color=color, linewidth=linewidth)

    if return_plot:
        return ax
    else:
        if show_plot:
            plt.show()


def matplot_boundaries(entity,
                       linewidth=1,
                       color='red',
                       return_plot=False,
                       show_plot=True):
    """
    Plot a fracability Boundary entity using matplotlib.

    :param entity: Boundary entity to plot
    :param linewidth: Size of the lines as int
    :param color: General color of the lines as str
    :param return_plot: Bool. If true the plot is returned. By default, False
    :param show_plot: Bool. If true the plot is shown. By default, True
    :return: If return_plot is true a matplotlib axis is returned
    """

    if return_plot:
        ax = plt.gca()
    else:
        figure = plt.figure(num=f'Boundary plot')

        ax = plt.subplot(111)

    entity.entity_df.plot(ax=ax, color=color, linewidth=linewidth)

    if return_plot:
        return ax
    else:
        if show_plot:
            plt.show()


def matplot_frac_net(entity,
                     markersize=7,
                     fracture_linewidth=1,
                     boundary_linewidth=1,
                     fracture_color='black',
                     boundary_color='red',
                     color_set=False,
                     show_plot=True,
                     return_plot=False):
    """
    Plot a fracability FractureNetwork entity using matplotlib.

    :param entity: FractureNetwork entity to plot
    :param markersize: Size of the nodes
    :param fracture_linewidth: Size of the fracture lines.
    :param boundary_linewidth: Size of the boundary lines.
    :param fracture_color: Color of the fracture lines.
    :param boundary_color: Color of the boundary lines.
    :param color_set: Bool. If true the lines are based on the set values.
    :param return_plot: Bool. If true the plot is returned. By default, False
    :param show_plot: Bool. If true the plot is shown. By default, True
    :return: If return_plot is true a matplotlib axis is returned

    """
    figure = plt.figure(num=f'Fracture net plot')
    ax = plt.subplot(111)
    nodes = entity.nodes
    fractures = entity.fractures
    boundary = entity.boundaries

    if fractures is not None:
        matplot_fractures(fractures, linewidth=fracture_linewidth,
                          color=fracture_color, color_set=color_set, return_plot=True)
    if boundary is not None:
        matplot_boundaries(boundary, linewidth=boundary_linewidth,
                           color=boundary_color, return_plot=True)
    if nodes is not None:
        matplot_nodes(nodes, markersize=markersize, return_plot=True)

    if return_plot:
        return ax
    else:
        if show_plot:
            plt.show()


# todo implement backbone plot with matplotlib
def matplot_backbone(entity,
                     fracture_linewidth=1,
                     boundary_linewidth=1,
                     fracture_color='black',
                     boundary_color='red',
                     color_set=False,
                     show_plot=True,
                     return_plot=False):
    """
    Plot a fracability FractureNetwork entity using matplotlib. (not yet implemented)

    :param entity: FractureNetwork entity to plot
    :param fracture_linewidth: Size of the fracture lines.
    :param boundary_linewidth: Size of the boundary lines.
    :param fracture_color: Color of the fracture lines.
    :param boundary_color: Color of the boundary lines.
    :param color_set: Bool. If true the lines are based on the set values.
    :param return_plot: Bool. If true the plot is returned. By default, False
    :param show_plot: Bool. If true the plot is shown. By default, True
    :return: If return_plot is true a matplotlib axis is returned

    """
    figure = plt.figure(num=f'Backbone plot')
    ax = plt.subplot(111)

    if 'backbone' in entity.entity_df['type'].values:
        backbone = entity.backbone
    else:
        entity.calculate_backbone()
        backbone = entity.backbone

    boundary = entity.boundaries

    if backbone is not None:
        for bb in backbone:
            matplot_fractures(bb, linewidth=fracture_linewidth,
                              color=fracture_color, color_set=color_set, return_plot=True)
    if boundary is not None:
        matplot_boundaries(boundary, linewidth=boundary_linewidth,
                           color=boundary_color, return_plot=True)

    if return_plot:
        return ax
    else:
        if show_plot:
            plt.show()


def vtkplot_nodes(entity,
                  markersize=7,
                  return_plot=False,
                  show_plot=True,
                  notebook=True):
    """
    Plot a fracability Nodes entity using vtk.

    :param entity: FractureNetwork entity to plot
    :param markersize: Size of the nodes
    :param return_plot: Bool. If true the plot is returned. By default, False
    :param show_plot: Bool. If true the plot is shown. By default, True
    :param notebook: Bool. if true plot using jupyter. By default, True
    :return: If return_plot is true a matplotlib axis is returned

    """
    plotter = Plotter(notebook=notebook)
    plotter.background_color = 'white'
    plotter.view_xy()
    plotter.add_camera_orientation_widget()
    plotter.enable_image_style()

    class_dict = {
        1: 'I',
        2: 'V',
        3: 'Y',
        4: 'X',
        5: 'U',
        6: 'Y2'
    }
    cmap_dict = {
        'I': 'Blue',
        'Y': 'Green',
        'Y2': 'Cyan',
        'X': 'Red',
        'U': 'Yellow'
    }

    nodes = entity.vtk_object

    class_names = [class_dict[int(i)] for i in nodes['n_type']]

    used_tags = list(set(class_names))
    used_tags.sort()
    cmap = [cmap_dict[i] for i in used_tags]

    sargs = dict(interactive=False,
                 vertical=False,
                 height=0.1,
                 title_font_size=16,
                 label_font_size=14)

    actor = plotter.add_mesh(nodes,
                             scalars=class_names,
                             render_points_as_spheres=True,
                             point_size=markersize,
                             show_scalar_bar=True,
                             scalar_bar_args=sargs,
                             cmap=cmap)

    if return_plot:
        return actor
    else:
        if show_plot:
            plotter.reset_camera()
            plotter.show()


def vtkplot_fractures(entity,
                      linewidth=1,
                      color='black',
                      color_set=False,
                      return_plot=False,
                      show_plot=True,
                      display_property: str = None,
                      notebook=True):

    """
    Plot a fracability Fracture entity using vtk.

    :param entity: Fracture entity to plot
    :param linewidth: width of the lines
    :param color: General color of the lines as str.
    :param color_set: Bool. If true the fractures are colored using the set.
    :param return_plot: Bool. If true the plot is returned. By default, False
    :param show_plot: Bool. If true the plot is shown. By default, True
    :param display_property: str. Indicate which property to show. By default, None
    :param notebook: Bool. if true plot using jupyter. By default, True

    :return: If return_plot is true a matplotlib axis is returned

    """

    plotter = Plotter(notebook=notebook)
    plotter.background_color = 'white'
    plotter.view_xy()
    plotter.add_camera_orientation_widget()
    plotter.enable_image_style()

    vtk_object = entity.vtk_object

    if color_set:
        display_property = 'f_set'

    if display_property:

        if display_property in vtk_object.array_names:  # this needs fixin'. Connecting fractures color are "contaminated"
            n_sets = len(set(vtk_object[display_property]))
            cmap = matplotlib.colormaps.get_cmap("rainbow").resampled(n_sets)
            actor = plotter.add_mesh(vtk_object,
                                     scalars=display_property,
                                     line_width=linewidth,
                                     cmap=cmap,
                                     show_scalar_bar=False)
        else:
            return False
    else:
        actor = plotter.add_mesh(vtk_object,
                                 color=color,
                                 line_width=linewidth,
                                 show_scalar_bar=False)

    if return_plot:
        return actor
    else:
        if show_plot:
            plotter.reset_camera()
            plotter.show()


def vtkplot_boundaries(entity,
                       linewidth=1,
                       color='red',
                       return_plot=False,
                       show_plot=True,
                       notebook=True):
    """
    Plot a fracability Boundary entity using vtk.

    :param entity: Fracture entity to plot
    :param linewidth: width of the lines
    :param color: General color of the lines as str.
    :param return_plot: Bool. If true the plot is returned. By default, False
    :param show_plot: Bool. If true the plot is shown. By default, True
    :param notebook: Bool. if true plot using jupyter. By default, True

    :return: If return_plot is true a matplotlib axis is returned
    """
    plotter = Plotter(notebook=notebook)
    plotter.background_color = 'white'
    plotter.view_xy()
    plotter.add_camera_orientation_widget()
    plotter.enable_image_style()

    actor = plotter.add_mesh(entity.vtk_object,
                     color=color,
                     line_width=linewidth,
                     show_scalar_bar=False)

    if return_plot:
        return actor
    else:
        if show_plot:
            plotter.reset_camera()
            plotter.show()


def vtkplot_frac_net(entity,
                     markersize=7,
                     fracture_linewidth=1,
                     boundary_linewidth=1,
                     fracture_color='black',
                     boundary_color='red',
                     color_set=False,
                     show_plot=True,
                     return_plot=False,
                     notebook=True):
    """
    Plot a fracability FractureNetwork entity using Pyvista.



    :param entity: FractureNetwork entity to plot
    :param markersize: Size of the nodes
    :param fracture_linewidth: Size of the fracture lines.
    :param boundary_linewidth: Size of the boundary lines.
    :param fracture_color: Color of the fracture lines.
    :param boundary_color: Color of the boundary lines.
    :param color_set: Bool. If true the lines are based on the set values.
    :param return_plot: Bool. If true the plot is returned. By default, False
    :param show_plot: Bool. If true the plot is shown. By default, True
    :param notebook: Bool. if true plot using jupyter. By default, True

    :return: If return_plot is true a matplotlib axis is returned

    """
    plotter = Plotter(notebook=notebook)
    plotter.background_color = 'white'
    plotter.view_xy()
    plotter.add_camera_orientation_widget()
    plotter.enable_image_style()

    nodes = entity.nodes
    fractures = entity.fractures
    boundaries = entity.boundaries

    if nodes is not None:
        node_actor = vtkplot_nodes(nodes, markersize=markersize, return_plot=True)
        plotter.add_actor(node_actor)

    if fractures is not None:
        fractures_actor = vtkplot_fractures(fractures, linewidth=fracture_linewidth,
                                            color=fracture_color, color_set=color_set, return_plot=True)
        plotter.add_actor(fractures_actor)

    if boundaries is not None:
        boundary_actor = vtkplot_boundaries(boundaries, linewidth=boundary_linewidth,
                                            color=boundary_color, return_plot=True)
        plotter.add_actor(boundary_actor)

    if return_plot:
        actors = plotter.actors
        return actors
    else:
        if show_plot:
            plotter.reset_camera()
            plotter.show()


def vtkplot_backbone(entity,
                     fracture_linewidth=1,
                     boundary_linewidth=1,
                     fracture_color='black',
                     boundary_color='red',
                     return_plot=False,
                     show_plot=True,
                     notebook=True):
    """
    Plot a fracability FractureNetwork backbone entity using pyvista.

    :param entity: FractureNetwork entity to plot
    :param markersize: Size of the nodes
    :param fracture_linewidth: Size of the fracture lines.
    :param boundary_linewidth: Size of the boundary lines.
    :param fracture_color: Color of the fracture lines.
    :param boundary_color: Color of the boundary lines.
    :param color_set: Bool. If true the lines are based on the set values.
    :param return_plot: Bool. If true the plot is returned. By default, False
    :param show_plot: Bool. If true the plot is shown. By default, True
    :param notebook: Bool. if true plot using jupyter. By default, True

    :return: If return_plot is true a matplotlib axis is returned

    """
    plotter = Plotter(notebook=notebook)
    plotter.background_color = 'white'
    plotter.view_xy()
    plotter.add_camera_orientation_widget()
    plotter.enable_image_style()

    if 'backbone' in entity.entity_df['type'].values:
        backbone = entity.backbone
    else:
        entity.calculate_backbone()
        backbone = entity.backbone

    boundaries = entity.boundaries

    if backbone is not None:
        for bb in backbone:
            backbone_actor = vtkplot_fractures(bb, linewidth=fracture_linewidth,
                                               color=fracture_color, return_plot=True)
            plotter.add_actor(backbone_actor)
    if boundaries is not None:
        boundary_actor = vtkplot_boundaries(boundaries, linewidth=boundary_linewidth,
                                            color=boundary_color, return_plot=True)
        plotter.add_actor(boundary_actor)

    if return_plot:
        actors = plotter.actors
        return actors
    else:
        if show_plot:
            plotter.reset_camera()
            plotter.show()


def matplot_stats_pdf(network_distribution: NetworkDistribution,
                      histogram: bool = True,
                      show_plot: bool = True):
    """
    Plot PDF and histogram.

    Parameters
    -------
    network_distribution: Input NetworkDistribution object

    histogram: Bool. If true plot also the histogram of the data. By default, is True

    show_plot: Bool. If true show the plot. By default, is True

    """

    distribution = network_distribution.distribution
    name = network_distribution.distribution_name

    network_data = network_distribution.fit_data

    x_vals = network_data.lengths

    if show_plot:
        fig = plt.figure(num=f'{name} PDF plot', figsize=(13, 7))
        fig.text(0.5, 0.95, name, ha='center')

    y_vals = distribution.pdf(x_vals)

    plt.plot(x_vals, y_vals, color='r', label=f'{name} pdf')
    if histogram:
        sns.histplot(x_vals, stat='density')

    plt.xlabel('length [m]')
    plt.title('PDF')
    plt.grid(False)
    plt.legend()

    if show_plot:
        plt.show()


def matplot_stats_cdf(network_distribution: NetworkDistribution,
                      plot_ecdf: bool = True,
                      show_plot: bool = True):
    """
    Plot CDF and ECDF.

    Parameters
    -------
    network_distribution: Input NetworkDistribution object

    plot_ecdf: Bool. If true plot also the empirical CDF curve of the data. By default, is True

    show_plot: Bool. If true show the plot. By default, is True

    """

    distribution = network_distribution.distribution
    name = network_distribution.distribution_name

    network_data = network_distribution.fit_data

    x_vals = network_data.lengths

    if show_plot:
        fig = plt.figure(num=f'{name} CDF plot', figsize=(13, 7))
        fig.text(0.5, 0.95, name, ha='center')

    y_vals = distribution.cdf(x_vals)

    plt.plot(x_vals, y_vals, color='r', label=f'{name} CDF')
    if plot_ecdf:
        ecdf = network_data.ecdf
        plt.step(x=x_vals, y=ecdf, color='b', label='Empirical CDF')

    plt.xlabel('length [m]')
    plt.title('CDF')
    plt.grid(True)
    plt.legend()

    if show_plot:
        plt.show()


def matplot_stats_sf(network_distribution: NetworkDistribution,
                     plot_esf: bool = True,
                     show_plot: bool = True):
    """
    Plot SF and ESF.

    Parameters
    -------
    network_distribution: Input NetworkDistribution object

    plot_esf: Bool. If true plot also the empirical SF curve of the data. By default, is True

    show_plot: Bool. If true show the plot. By default, is True

    """
    distribution = network_distribution.distribution
    name = network_distribution.distribution_name

    network_data = network_distribution.fit_data

    x_vals = network_data.lengths

    if show_plot:
        fig = plt.figure(num=f'{name} SF plot', figsize=(13, 7))
        fig.text(0.5, 0.95, name, ha='center')

    y_vals = distribution.sf(x_vals)

    plt.plot(x_vals, y_vals, color='r', label=f'{name} SF')
    if plot_esf:
        esf = network_data.esf
        plt.step(x=x_vals, y=esf, color='b', label='Empirical SF')

    plt.xlabel('length [m]')
    plt.title('SF')
    plt.grid(True)
    plt.legend()

    if show_plot:
        plt.show()


def matplot_stats_table(network_distribution: NetworkDistribution,
                        vertical: bool = True,
                        show_plot: bool = True):
    """
    Plot the stats summary table for both the data and the NetworkDistribution. In particular the following
    estimators are calculated:

        1. Mean
        2. Standard Deviation
        3. Variance
        4. Median
        5. Mode
        6. 5th percentile
        7. 95th percentile
        8. Total number of fractures
        9. % censored

    Parameters
    -----------
    network_distribution: Input NetworkDistribution object

    vertical: Bool. If true the table is vertical (2cols x 7rows). By default, is True

    show_plot: Bool. If true show the plot. By default, is True

    """
    name = network_distribution.distribution_name
    if show_plot:
        fig = plt.figure(num=f'{name} summary table')
        fig.text(0.5, 0.95, f'{name} summary table', ha='center')

    network_data = network_distribution.fit_data
    plt.axis("off")
    dec = 4

    text_mean = f'{np.round(network_distribution.mean, dec)}'
    text_std = f'{np.round(network_distribution.std, dec)}'
    text_var = f'{np.round(network_distribution.var, dec)}'
    text_median = f'{np.round(network_distribution.median, dec)}'
    text_mode = f'{np.round(network_distribution.mode[0], dec)}'
    text_b5 = f'{np.round(network_distribution.b5, dec)}'
    text_b95 = f'{np.round(network_distribution.b95, dec)}'

    text_mean_th = f'{np.round(network_data.mean, dec)}'
    text_std_th = f'{np.round(network_data.std, dec)}'
    text_var_th = f'{np.round(network_data.var, dec)}'
    text_median_th = f'{np.round(network_data.median, dec)}'
    text_mode_th = f'{np.round(network_data.mode[0], dec)}'
    text_b5_th = f'{np.round(network_data.b5, dec)}'
    text_b95_th = f'{np.round(network_data.b95, dec)}'

    text_totalf = f'{network_data.total_n_fractures}'
    text_perc_censor = f'{np.round(network_data.censoring_percentage, dec)}'

    text_AIC_rank = f'{network_distribution.Akaike_rank}'
    text_KS_rank = f'{network_distribution.KS_rank}'
    text_KG_rank = f'{network_distribution.KG_rank}'
    text_AD_rank = f'{network_distribution.AD_rank}'
    text_mean_rank = f'{network_distribution.Mean_rank}'

    general_stats_df = pd.DataFrame(data=[[text_totalf, text_perc_censor]],
                                    columns=['Total number of fractures', '% censored'])

    stats_df = pd.DataFrame(data=[[text_mean_th, text_mean],
                                  [text_median_th, text_median],
                                  [text_mode_th, text_mode],
                                  [text_b5_th, text_b5],
                                  [text_b95_th, text_b95],
                                  [text_std_th, text_std],
                                  [text_var_th, text_var]],
                            columns=['Data', 'Fit'], index=['Mean', 'Median', 'Mode',
                                                            'B5', 'B95', 'Std', 'Var'])

    rank_df = pd.DataFrame(data=[[text_AIC_rank,text_KS_rank, text_KG_rank, text_AD_rank, text_mean_rank]],
                           columns=['Akaike rank', 'KS rank', 'KG rank', 'AD rank', 'Mean rank'])

    if not vertical:
        stats_df = stats_df.transpose()

    gen_table = plt.table(cellText=general_stats_df.values,
                          colLabels=general_stats_df.columns,
                          # colWidths=[0.3, 0.3],
                          loc='upper center',cellLoc='center')

    # gen_table.auto_set_font_size(False)
    # gen_table.auto_set_column_width(col=list(range(len(general_stats_df.columns))))

    sta_table = plt.table(cellText=stats_df.values,
                          rowLabels=stats_df.index,
                          colLabels=stats_df.columns,
                          # colWidths=[0.3, 0.3],
                          loc='center')

    # sta_table.auto_set_font_size(False)
    # sta_table.auto_set_column_width(col=list(range(len(stats_df.columns))))

    ran_table = plt.table(cellText=rank_df.values,
                          colLabels=rank_df.columns,
                          # colWidths=[0.3, 0.3],
                          loc='lower center', cellLoc='center')

    # ran_table.auto_set_font_size(False)
    # ran_table.auto_set_column_width(col=list(range(len(rank_df.columns))))

    if show_plot:
        plt.show()


def matplot_stats_summary(fitter: NetworkFitter,
                          function_list: list = ['pdf', 'cdf', 'sf'],
                          table: bool = True,
                          show_plot: bool = True,
                          position: list = [],
                          sort_by: str = 'Akaike'):
    """
    Summarize PDF, CDF and SF functions and display summary table all
    in a single plot.


    Parameters
    -------
    fitter: Input NetworkFitter object

    function_list: List of function to calculate (cdf, pdf etc.). By default pdf, cdf and sf functions are calculated

    table: Bool. If true the summary table is shown. By default is true

    show_plot: Bool. If true show the plot. By default, is True
    
    position: List.  Plot the models at the given position (1 indexed) in the ordered sorted by the sort_by field. If None show the plots for each fit. By default is [1].

    sort_by: str. If best is True, show the best fit using the indicated column name. By default is Akaike

    """

    # This distribution selection process could be probably optimized
    names = []
    if position:
        for pos in position:
            names.append(fitter.get_fitted_distribution_names(sort_by)[pos-1])
    else:
        names = fitter.get_fitted_distribution_names(sort_by)

    for name in names:
        network_distribution = fitter.get_fitted_distribution(name)
        fig = plt.figure(num=f'{name} summary plot', figsize=(13, 7))
        # fig.text(0.5, 0.02, 'Length [m]', ha='center')
        fig.suptitle(name)
        # fig.text(0.04, 0.5, 'Density', va='center', rotation='vertical')

        for i, func_name in enumerate(function_list):

            if func_name == 'pdf':
                plt.subplot(2, 2, i+1)
                matplot_stats_pdf(network_distribution, show_plot=False)
            if func_name == 'cdf':
                plt.subplot(2, 2, i + 1)
                matplot_stats_cdf(network_distribution, show_plot=False)
            if func_name == 'sf':
                plt.subplot(2, 2, i + 1)
                matplot_stats_sf(network_distribution, show_plot=False)
        if table:
            plt.subplot(2, 2, i+2)
            plt.axis("off")
            plt.ylim([0, 8])
            plt.xlim([0, 10])
            matplot_stats_table(network_distribution, show_plot=False)

            plt.tight_layout()

        if show_plot:
            plt.show()


def matplot_stats_uniform(fitter: NetworkFitter,
                          show_plot: bool = True,
                          position: list = None,
                          sort_by: str = 'Akaike',
                          bw: bool = False,
                          second_axis: bool = True,
                          n_ticks: int = 5):
    """
    Confront the fitted data with the standard uniform 0,1. Following Kim 2019

    Parameters
    -------
    fitter: Input NetworkFitter object

    show_plot: Bool. If true show the plot. By default, is True

    position: List.  Plot the models at the given position (1 indexed) in the ordered fit_records dataframe sorted by the sort_by field. If None show the plots for each fit. Default is None

    sort_by: str. If best is True, show the best fit using the indicated column name. By default is Akaike

    bw: Bool. If true turn the plot color-blind friendly (black lines with different patterns). Max 7 lines (i.e. models). Default is False

    second_axis: Bool. If true plot the secondary X axis representing the corresponding values of the cumulative in meters. Default is True

    n_ticks: int. Number of ticks on the second axis.

    todo add custom second axis values
    """

    delta = fitter.network_data.delta
    ecdf = fitter.network_data.ecdf
    samples = fitter.network_data.lengths

    # This distribution selection process could be probably optimized
    names = []
    if position:
        for pos in position:
            names.append(fitter.get_fitted_distribution_names(sort_by)[pos-1])
    else:
        names = fitter.get_fitted_distribution_names(sort_by)

    fig = plt.figure(num=f'Comparison plot', figsize=(13, 7))
    ax = plt.subplot(111)

    for name in names:
        network_distribution = fitter.get_fitted_distribution(name)
        Z = network_distribution.cdf()
        G_n = KM(Z, Z, delta)  # not sure about this!
        ax.plot(Z, G_n, label=f'{name}')

    if bw:
        setFigLinesBW(fig)

    if second_axis:
        x_ticks = np.linspace(0, 1, n_ticks)

        xticks_labels = ecdf_find_x(samples=samples, ecdf_prob=ecdf, y_values=x_ticks)

        ax2 = ax.twiny()
        ax2.plot(x_ticks, [0] * len(x_ticks), visible=False)
        ax2.spines['top'].set_position(('axes', -0.11))
        ax2.spines['top'].set_visible(False)
        ax2.spines['bottom'].set_position(('axes', -0.05))
        ax2.spines['bottom'].set_visible(True)
        ax2.set_xticks(x_ticks)
        ax2.set_xticklabels(xticks_labels)

        ax2.annotate(f'Reference length [m]', (-0.01, 0.5), ha='right', va='center', xycoords=ax2.spines['bottom'])

        plt.tick_params(which='both', direction='inout', bottom=True, top=False, length=6)

    if show_plot:
        ax.plot([0, 1], [0, 1], color='r', label='U (0,1)')
        ax.grid(False)
        ax.legend()
        plt.title('Distance to Uniform comparison')
        plt.show()


def matplot_tick_plot(fitter: NetworkFitter,
                      show_plot: bool = True,
                      position: list = None,
                      n_ticks: int = 5,
                      sort_by: str = 'Akaike'):
    """
    Plot the tick plot of the fitted models. (Describe the tick plot).
    :param position:
    :param n_ticks:
    :param fitter:
    :param show_plot:
    :param sort_by:
    :return:
    """

    names = []
    if position:
        for pos in position:
            names.append(fitter.get_fitted_distribution_names(sort_by)[pos-1])
    else:
        names = fitter.get_fitted_distribution_names(sort_by)

    fig = plt.figure(num=f'Tick plot', figsize=(13, 7))
    ax = plt.subplot(111)

    figManager = plt.get_current_fig_manager()
    figManager.window.showMaximized()

    samples = fitter.network_data.lengths
    ecdf = fitter.network_data.ecdf

    x_ticks = np.linspace(0, 1, n_ticks)

    xticks_labels = ecdf_find_x(samples=samples, ecdf_prob=ecdf, y_values=x_ticks)

    for i, name in enumerate(names):
        network_distribution = fitter.get_fitted_distribution(name)
        y = i
        plt.axhline(y=y, color='k')
        values_labels = network_distribution.cdf(xticks_labels)
        for value, x in zip(values_labels, xticks_labels):
            point = ax.plot(value, y, 'k|', markersize=12, linewidth=1)
            ax.annotate(xy=(0.5, -1.1), text=f'{x}', ha='center', va='center', xycoords=point[0], annotation_clip=True)

    ax.set_xlim(-0.1, 1.1)
    ax.set_ylim(bottom=-0.5)
    ax.set_yticks(range(len(names)))
    ax.set_yticklabels(names)
    ax.set_xticks(x_ticks)
    ax.set_xticklabels(xticks_labels)
    ax.set_xlabel('Length[m]')
    ax.annotate(f'Kaplan-Meier', (-0.01, 0.5), ha='right', va='center', xycoords=ax.spines['bottom'])

    ax.grid('y')
    ax.spines[['left', 'top', 'right']].set_visible(False)

    if show_plot:
        plt.title('Tick plot')
        plt.show()